#!/usr/bin/env python

#****************************************************************************************************************************************************
# Copyright (c) 2014 Freescale Semiconductor, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
#    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
#      its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#****************************************************************************************************************************************************

from FslBuildGen.DataTypes import *
from FslBuildGen.Exceptions import *
from FslBuildGen.XmlStuff import XmlGenFile
from FslBuildGen import IOUtil
from FslBuildGen.PackageTemplateLoader import PackageTemplateLoader
import os


class PackageLoader(object):
    def __init__(self, config, files):
        super(PackageLoader, self).__init__()

        templateLocationCache = self.__CacheTemplateLocations(config)
        self.PackageTemplateLoader = PackageTemplateLoader(templateLocationCache)

        config.LogPrint("- Parsing")

        if config.IsSDKBuild:
            files = self.__PreloadPackages(config, files)

        packageDict = {}
        genFiles = []
        # Load the initial package files
        self.__LoadFiles(config, files, packageDict, genFiles)
        
        locationCache = self.__CachePackageLocations(config)

        searchForPackages = True
        newGenFiles = genFiles
        while searchForPackages:
            missingPackages = self.__DiscoverMissingPackages(packageDict, newGenFiles)
            if len(missingPackages) > 0:
                newFiles = self.__LocateMissingPackages(locationCache, missingPackages, config.GenFileName)
                newGenFiles = []
                self.__LoadFiles(config, newFiles, packageDict, newGenFiles)
                genFiles += newGenFiles
            else:
                searchForPackages = False

        self.GenFiles = genFiles


    
    def __CacheTemplateLocations(self, config):
        """ Build a dict of all *.gen files found in the template import directories """
        dict = {}
        for location in config.TemplateImportDirectories:
            files = IOUtil.GetFilesAt(location.ResolvedPath, True)
            for entry in files:
                if entry.endswith(".gen"):
                    dict[IOUtil.GetFileNameWithoutExtension(entry)] = IOUtil.Join(location.ResolvedPath, entry)
        return dict


    def __CachePackageLocations(self, config):
        dict = {}
        packageConfiguration = config.ToolConfig.PackageConfiguration[config.Type]
        for location in packageConfiguration.Locations:
            dirs = IOUtil.GetDirectoriesAt(location.ResolvedPath, False)
            for entry in dirs:
                fullPath = IOUtil.Join(location.ResolvedPath, entry)
                if not self.__IsBlackListed(location, fullPath):
                    dict[entry] = fullPath
        return dict


    def __LocateMissingPackages(self, locationCache, missing, genFileName):
        files = []
        for entry in sorted(missing.keys()):
            found = False
            if entry in locationCache:
                fullPathPackageFile = IOUtil.Join(locationCache[entry], genFileName)
                if os.path.isfile(fullPathPackageFile):
                    files.append(fullPathPackageFile)
                    found = True
            if not found:
                raise DependencyNotFoundException(missing[entry], entry)
        return files


    def __PreloadPackages(self, config, theFiles):
        # Get all the files associated with the typeId then merge it with the supplied file list
        files = self.__GetFiles(config) 
        for theFile in theFiles:
            if not theFile in files:
                files.append(theFile)
        files.sort()
        return files


    def __IsBlackListed(self, location, path):
        for entry in location.Blacklist:
            entryName = entry.Name + '/'
            if path.startswith(entryName):
                return True
        return False

    def __LocatePackages(self, location, genFileName):
        files = []
        dirs = IOUtil.GetDirectoriesAt(location.ResolvedPath, False)
        for dir in dirs:
            packageFile = IOUtil.Join(dir, genFileName)
            fullPathPackageFile = IOUtil.Join(location.ResolvedPath, packageFile)
            if os.path.isfile(fullPathPackageFile) and not self.__IsBlackListed(location, packageFile):
                files.append(fullPathPackageFile)
        return files


    def __GetFiles(self, config):
        packageConfiguration = config.ToolConfig.PackageConfiguration[config.Type]
        files = []
        for location in packageConfiguration.Locations:
            files += self.__LocatePackages(location, config.GenFileName)
        return files


    def __DiscoverMissingPackages(self, packageDict, genFiles):
        missingPackages = {}
        for entry in genFiles:
            for dep in entry.DirectDependencies:
                if not dep.Name in packageDict and not dep.Name in missingPackages:
                    missingPackages[dep.Name] = entry
            for platform in entry.Platforms.values():
                for dep in platform.DirectDependencies:
                    if not dep.Name in packageDict and not dep.Name in missingPackages:
                        missingPackages[dep.Name] = entry
        return missingPackages


    def __LoadFiles(self, config, files, packageDict, genFiles):
        for file in files:
            config.LogPrint("  '%s'" % (file))
            xml = XmlGenFile()
            xml.Load(config, self.PackageTemplateLoader, file)
            genFiles.insert(0, xml)
            if not packageDict.has_key(xml.Name):
                packageDict[xml.Name] = [xml]
            else:
                packageDict[xml.Name].append(xml)

        for packageList in packageDict.values():
            if len(packageList) > 1:
                raise PackageHasMultipleDefinitionsException(packageList)
