/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/BasicTypes.hpp>
#include <FslBase/Bits/ByteArrayUtil.hpp>
#include <BasicDDSLoader/BasicDDSLoader.hpp>
#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/RawBitmapEx.hpp>
#include <FslGraphics/RawBitmapUtil.hpp>
#include <cassert>
#include <algorithm>

namespace Fsl
{
  // Information added while trying to fix the basic loader just enough for it to work
  // for our input. It's really not recommended to use it for anything else

  // https://msdn.microsoft.com/en-us/library/windows/apps/jj651550.aspx
  // 
  // 0     4bytes magic number followed by DDS_HEADER
  //
  //       typedef struct
  //       {
  //  1      uint32          size;   
  //  2      uint32          flags;
  //  3      uint32          height;
  //  4      uint32          width;
  //  5      uint32          pitchOrLinearSize;
  //  6      uint32          depth; // only if DDS_HEADER_FLAGS_VOLUME is set in flags
  //  7      uint32          mipMapCount;
  //  8      uint32          reserved1[11];
  //  9      DDS_PIXELFORMAT ddspf;
  //  18     uint32          caps;
  //  19     uint32          caps2;
  //  20     uint32          caps3;
  //  21     uint32          caps4;
  //  22     uint32          reserved2;
  //       } DDS_HEADER;

  // Beware there is a DDS_HEADER_DXT10 as well

  //       struct DDS_PIXELFORMAT
  //       {
  //  10      uint32  size;
  //  11     uint32  flags;
  //  12     uint32  fourCC;
  //  13     uint32  RGBBitCount;
  //  14     uint32  RBitMask;
  //  15     uint32  GBitMask;
  //  16     uint32  BBitMask;
  //  17     uint32  ABitMask;
  //       };


  RawCubeBitmap BasicDDSLoader::Extract(uint8_t* pRawData, const int32_t length)
  {
    assert(pRawData != nullptr);
    assert(length >= 0);


    // read first 4 bytes, magic number
    const uint32_t magic = ByteArrayUtil::ReadUInt32LE(pRawData, length, (4 * 0));
    // check magic number, should be "DDS ": 0x20534444.
    if (magic != 0x20534444UL)
      throw GraphicsException("Invalid DDS data");
    
    // second 4 bytes: texture offset in file.
    const uint32_t headerSize = ByteArrayUtil::ReadUInt32LE(pRawData, length, (4 * 1));
    const uint32_t texWidth = ByteArrayUtil::ReadUInt32LE(pRawData, length, (4 * 3));
    const uint32_t texHeight = ByteArrayUtil::ReadUInt32LE(pRawData, length, (4 * 4));
    const uint32_t mipmapCount = std::max(ByteArrayUtil::ReadUInt32LE(pRawData, length, (4 * 7)), uint32_t(1));

    uint32_t bytesPerBitmap = 0;
    // fast way to calc inspired by the way Microsoft reads it
    {
      uint32_t w = texWidth;
      uint32_t h = texHeight;
      for (uint32_t i = 0; i < mipmapCount; ++i)
      {
        bytesPerBitmap += w * h * 4;
        w = w >> 1;
        h = h >> 1;
        if (w == 0)
          w = 1;
        if (h == 0)
          h = 1;
      }
    }

    uint32_t offset = sizeof(uint32_t) + headerSize;

    //load all six faces
    RawBitmapEx rawBitmapPosX(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);
    offset += bytesPerBitmap;

    RawBitmapEx rawBitmapNegX(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);
    offset += bytesPerBitmap;

    RawBitmapEx rawBitmapPosY(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);
    offset += bytesPerBitmap;

    RawBitmapEx rawBitmapNegY(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);
    offset += bytesPerBitmap;

    RawBitmapEx rawBitmapPosZ(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);
    offset += bytesPerBitmap;

    RawBitmapEx rawBitmapNegZ(pRawData + offset, texWidth, texHeight, PixelFormat::RGBA8888);

    // Fix the color layout
    RawBitmapUtil::Swizzle32(rawBitmapPosX, 2, 1, 0, 3);
    RawBitmapUtil::Swizzle32(rawBitmapNegX, 2, 1, 0, 3);
    RawBitmapUtil::Swizzle32(rawBitmapPosY, 2, 1, 0, 3);
    RawBitmapUtil::Swizzle32(rawBitmapNegY, 2, 1, 0, 3);
    RawBitmapUtil::Swizzle32(rawBitmapPosZ, 2, 1, 0, 3);
    RawBitmapUtil::Swizzle32(rawBitmapNegZ, 2, 1, 0, 3);

    return RawCubeBitmap(rawBitmapPosX, rawBitmapNegX, rawBitmapPosY, rawBitmapNegY, rawBitmapPosZ, rawBitmapNegZ);
  }

}


