#ifndef FSLBASE_MATH_RECTANGLE_HPP
#define FSLBASE_MATH_RECTANGLE_HPP
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <algorithm>
#include <FslBase/Math/Point2.hpp>
//#include <FslBase/Math/Vector2.hpp>

namespace Fsl
{
  struct Rectangle
  {
    int X;
    int Y;
    int Width;
    int Height;

    Rectangle()
      : X(0)
      , Y(0)
      , Width(0)
      , Height(0)
    {
    }


    Rectangle(const int x, const int y, const int width, const int height)
      : X(x)
      , Y(y)
      , Width(width)
      , Height(height)
    {
      //assert(IsValid());
    }


    static Rectangle Empty() { return Rectangle(0, 0, 0, 0); }

    //! @brief Check if the rectangle is considered t be valid
    bool IsValid() const
    {
      return Width >= 0 && Height >= 0;
    }

    int Left() const { return X; }
    int Top() const { return Y; }
    int Right() const { return X + Width; }
    int Bottom() const { return Y + Height; }
    
    inline Point2 GetSize() const { return Point2(Width, Height); }
    
    inline Point2 TopLeft() const { return Point2(X, Y); }
    inline Point2 TopRight() const { return Point2(Right(), Y); }
    inline Point2 BottomLeft() const { return Point2(X, Bottom()); }
    inline Point2 BottomRight() const { return Point2(Right(), Bottom()); }


    //! @brief Check if the x,y coordinate is considered to be contained within this rectangle
    bool Contains(const int x, const int y) const
    {
      return ((((x >= X) && (x < (X + Width))) && (y >= Y)) && (y < (Y + Height)));
    }


    //! @brief Check if the x,y coordinate is considered to be contained within this rectangle
    bool Contains(float x, float y) const
    {
      return ((((x >= x) && (x < (X + Width))) && (y >= Y)) && (y < (Y + Height)));
    }

    //! @brief Check if the x,y coordinate is considered to be contained within this rectangle
    bool Contains(const Point2& value) const
    {
      return ((((value.X >= X) && (value.X < (X + Width))) && (value.Y >= Y)) && (value.Y < (Y + Height)));
    }

    //! @brief Check if the x,y coordinate is considered to be contained within this rectangle
    //bool Contains(const Vector2& value) const
    //{
    //  return ((((value.X >= X) && (value.X < (X + Width))) && (value.Y >= Y)) && (value.Y < (Y + Height)));
    //}

    //! @brief Check if the rectangle is considered to be contained within this rectangle
    bool Contains(Rectangle value) const
    {
      return ((((value.X >= X) && ((value.X + value.Width) <= (X + Width))) &&
                (value.Y >= Y)) && ((value.Y + value.Height) <= (Y + Height)));
    }


    //! @brief Get the start location of this rect
    const Point2 GetLocation() const
    {
      return Point2(X, Y);
    }


    //! @brief Set the start location of this rect
    void SetLocation(const Point2& location)
    {
      X = location.X;
      Y = location.Y;
    }


    //! @brief Get the center of this rect
    Point2 GetCenter() const
    {
      return Point2(X + (Width / 2), Y + (Height / 2));
    }


    //! @brief Each corner of the Rectangle is pushed away from the center of the rectangle by the specified amounts.
    //!        This results in the width and height of the Rectangle increasing by twice the values provide
    void Inflate(const int horizontalValue, const int verticalValue)
    {
      X -= horizontalValue;
      Y -= verticalValue;
      Width += horizontalValue * 2;
      Height += verticalValue * 2;
    }


    //! @brief Gets a value that indicates whether the Rectangle is empty
    //!        An empty rectangle has all its values set to 0.
    bool IsEmpty() const
    {
      return ((((Width == 0) && (Height == 0)) && (X == 0)) && (Y == 0));
    }


    //! @brief Determines whether a specified Rectangle intersects with this Rectangle.
    bool Intersects(const Rectangle& value) const
    {
      return value.Left() < Right() && Left() < value.Right() && value.Top() < Bottom() && Top() < value.Bottom();
    }


    //! @brief Creates a Rectangle defining the area where one rectangle overlaps another rectangle.
    static Rectangle Intersect(const Rectangle& rect1, const Rectangle& rect2)
    {
      if (rect1.Intersects(rect2))
      {
        int rightSide = std::min(rect1.X + rect1.Width, rect2.X + rect2.Width);
        int leftSide = std::max(rect1.X, rect2.X);
        int topSide = std::max(rect1.Y, rect2.Y);
        int bottomSide = std::min(rect1.Y + rect1.Height, rect2.Y + rect2.Height);
        return Rectangle(leftSide, topSide, rightSide - leftSide, bottomSide - topSide);
      }
      else
      {
        return Rectangle(0, 0, 0, 0);
      }
    }


    //! @brief Creates a new Rectangle that exactly contains the two supplied rectangles
    static Rectangle Union(const Rectangle& rect1, const Rectangle& rect2)
    {
      const int x = std::min(rect1.X, rect2.X);
      const int y = std::min(rect1.Y, rect2.Y);
      return Rectangle(x, y, std::max(rect1.Right(), rect2.Right()) - x, std::max(rect1.Bottom(), rect2.Bottom()) - y);
    }


    bool operator==(const Rectangle& rhs) const
    {
      return ((X == rhs.X) && (Y == rhs.Y) && (Width == rhs.Width) && (Height == rhs.Height));
    }


    bool operator!=(const Rectangle& rhs) const
    {
      return ((X != rhs.X) || (Y != rhs.Y) || (Width != rhs.Width) || (Height != rhs.Height));
    }

  };
}


#endif

