/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

// The functions in this file are a port of an MIT licensed library: MonaGame - Matrix.cs.

/*
MIT License
Copyright © 2006 The Mono.Xna Team

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <cassert>
#include <cmath>
#include <FslBase/Exceptions.hpp>
#include <FslBase/Math/Matrix.hpp>
//#include <FslBase/Math/Plane.hpp>
#include <FslBase/Math/Vector2.hpp>
#include <FslBase/Math/Vector4.hpp>
#include "MatrixInternals.hpp"

#define _M11 (4 * 0 + 0)
#define _M12 (4 * 0 + 1)
#define _M13 (4 * 0 + 2)
#define _M14 (4 * 0 + 3)
#define _M21 (4 * 1 + 0)
#define _M22 (4 * 1 + 1)
#define _M23 (4 * 1 + 2)
#define _M24 (4 * 1 + 3)
#define _M31 (4 * 2 + 0)
#define _M32 (4 * 2 + 1)
#define _M33 (4 * 2 + 2)
#define _M34 (4 * 2 + 3)
#define _M41 (4 * 3 + 0)
#define _M42 (4 * 3 + 1)
#define _M43 (4 * 3 + 2)
#define _M44 (4 * 3 + 3)

namespace Fsl
{
  Matrix::Matrix()
  {
    m[_M11] = 0.0f; m[_M12] = 0.0f; m[_M13] = 0.0f; m[_M14] = 0.0f;
    m[_M21] = 0.0f; m[_M22] = 0.0f; m[_M23] = 0.0f; m[_M24] = 0.0f;
    m[_M31] = 0.0f; m[_M32] = 0.0f; m[_M33] = 0.0f; m[_M34] = 0.0f;
    m[_M41] = 0.0f; m[_M42] = 0.0f; m[_M43] = 0.0f; m[_M44] = 0.0f;
  }


  Matrix::Matrix(const float m11, const float m12, const float m13, const float m14,
                 const float m21, const float m22, const float m23, const float m24,
                 const float m31, const float m32, const float m33, const float m34,
                 const float m41, const float m42, const float m43, const float m44)
  {
    m[_M11] = m11;
    m[_M12] = m12;
    m[_M13] = m13;
    m[_M14] = m14;
    m[_M21] = m21;
    m[_M22] = m22;
    m[_M23] = m23;
    m[_M24] = m24;
    m[_M31] = m31;
    m[_M32] = m32;
    m[_M33] = m33;
    m[_M34] = m34;
    m[_M41] = m41;
    m[_M42] = m42;
    m[_M43] = m43;
    m[_M44] = m44;
  }


  Vector3 Matrix::GetBackward() const
  {
    return Vector3(m[_M31], m[_M32], m[_M33]);
  }


  void Matrix::SetBackward(const Vector3& value)
  {
    m[_M31] = value.X;
    m[_M32] = value.Y;
    m[_M33] = value.Z;
  }


  Vector3 Matrix::GetDown() const
  {
    return Vector3(-m[_M21], -m[_M22], -m[_M23]);
  }


  void Matrix::SetDown(const Vector3& value)
  {
    m[_M21] = -value.X;
    m[_M22] = -value.Y;
    m[_M23] = -value.Z;
  }


  Vector3 Matrix::GetForward() const
  {
    return Vector3(-m[_M31], -m[_M32], -m[_M33]);
  }


  void Matrix::SetForward(const Vector3& value)
  {
    m[_M31] = -value.X;
    m[_M32] = -value.Y;
    m[_M33] = -value.Z;
  }


  Vector3 Matrix::GetLeft() const
  {
    return Vector3(-m[_M11], -m[_M12], -m[_M13]);
  }


  void Matrix::SetLeft(const Vector3& value)
  {
    m[_M11] = -value.X;
    m[_M12] = -value.Y;
    m[_M13] = -value.Z;
  }


  Vector3 Matrix::GetRight() const
  {
    return Vector3(m[_M11], m[_M12], m[_M13]);
  }


  void Matrix::SetRight(const Vector3& value)
  {
    m[_M11] = value.X;
    m[_M12] = value.Y;
    m[_M13] = value.Z;
  }


  Vector3 Matrix::GetTranslation() const
  {
    return Vector3(m[_M41], m[_M42], m[_M43]);
  }


  void Matrix::SetTranslation(const Vector3& value)
  {
    m[_M41] = value.X;
    m[_M42] = value.Y;
    m[_M43] = value.Z;
  }


  Vector3 Matrix::GetUp() const
  {
    return Vector3(m[_M21], m[_M22], m[_M23]);
  }


  void Matrix::SetUp(const Vector3& value)
  {
    m[_M21] = value.X;
    m[_M22] = value.Y;
    m[_M23] = value.Z;
  }


  void Matrix::Clear()
  {
    m[_M11] = 0.0f; m[_M12] = 0.0f; m[_M13] = 0.0f; m[_M14] = 0.0f;
    m[_M21] = 0.0f; m[_M22] = 0.0f; m[_M23] = 0.0f; m[_M24] = 0.0f;
    m[_M31] = 0.0f; m[_M32] = 0.0f; m[_M33] = 0.0f; m[_M34] = 0.0f;
    m[_M41] = 0.0f; m[_M42] = 0.0f; m[_M43] = 0.0f; m[_M44] = 0.0f;
  }


  Matrix Matrix::CreateLookAt(const Vector3& cameraPosition, const Vector3& cameraTarget, const Vector3& cameraUpVector)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateLookAt(result, cameraPosition, cameraTarget, cameraUpVector);
    return result;
  }


  void Matrix::CreateLookAt(Matrix& rResult, const Vector3& cameraPosition, const Vector3& cameraTarget, const Vector3& cameraUpVector)
  {
    const Vector3 vector = Vector3::Normalize(cameraPosition - cameraTarget);
    const Vector3 vector2 = Vector3::Normalize(Vector3::Cross(cameraUpVector, vector));
    const Vector3 vector3 = Vector3::Cross(vector, vector2);
    rResult.m[_M11] = vector2.X;
    rResult.m[_M12] = vector3.X;
    rResult.m[_M13] = vector.X;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = vector2.Y;
    rResult.m[_M22] = vector3.Y;
    rResult.m[_M23] = vector.Y;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = vector2.Z;
    rResult.m[_M32] = vector3.Z;
    rResult.m[_M33] = vector.Z;
    rResult.m[_M34] = 0.0f;
    rResult.m[_M41] = -Vector3::Dot(vector2, cameraPosition);
    rResult.m[_M42] = -Vector3::Dot(vector3, cameraPosition);
    rResult.m[_M43] = -Vector3::Dot(vector, cameraPosition);
    rResult.m[_M44] = 1.0f;
  }


  Matrix Matrix::CreateOrthographic(const float width, const float height, const float zNearPlane, const float zFarPlane)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateOrthographic(result, width, height, zNearPlane, zFarPlane);
    return result;
  }


  void Matrix::CreateOrthographic(Matrix& rResult, const float width, const float height, const float zNearPlane, const float zFarPlane)
  {
    if (width <= 0.0f)
      throw std::invalid_argument("width <= 0.0f");
    if (height <= 0.0f)
      throw std::invalid_argument("height <= 0.0f");

    rResult.m[_M11] = 2.0f / width;
    rResult.m[_M12] = 0.0f;
    rResult.m[_M13] = 0.0f;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = 0.0f;
    rResult.m[_M22] = 2.0f / height;
    rResult.m[_M23] = 0.0f;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = 0.0f;
    rResult.m[_M32] = 0.0f;
    rResult.m[_M33] = 1.0f / (zNearPlane - zFarPlane);
    rResult.m[_M34] = 0.0f;
    rResult.m[_M41] = 0.0f;
    rResult.m[_M42] = 0.0f;
    rResult.m[_M43] = zNearPlane / (zNearPlane - zFarPlane);
    rResult.m[_M44] = 1.0f;
  }


  Matrix Matrix::CreatePerspective(const float width, const float height, const float nearPlaneDistance, const float farPlaneDistance)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreatePerspective(result, width, height, nearPlaneDistance, farPlaneDistance);
    return result;
  }


  void Matrix::CreatePerspective(Matrix& rResult, const float width, const float height, const float nearPlaneDistance, const float farPlaneDistance)
  {
    if (width <= 0.0f)
      throw std::invalid_argument("width <= 0.0f");
    if (height <= 0.0f)
      throw std::invalid_argument("height <= 0.0f");
    if (nearPlaneDistance <= 0.0f)
      throw std::invalid_argument("nearPlaneDistance <= 0.0f");
    if (farPlaneDistance <= 0.0f)
      throw std::invalid_argument("farPlaneDistance <= 0.0f");
    if (nearPlaneDistance >= farPlaneDistance)
      throw std::invalid_argument("nearPlaneDistance >= farPlaneDistance");

    rResult.m[_M11] = (2.0f * nearPlaneDistance) / width;
    rResult.m[_M12] = 0.0f;
    rResult.m[_M13] = 0.0f;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = 0.0f;
    rResult.m[_M22] = (2.0f * nearPlaneDistance) / height;
    rResult.m[_M23] = 0.0f;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = 0.0f;
    rResult.m[_M32] = 0.0f;
    rResult.m[_M33] = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
    rResult.m[_M34] = -1.0f;
    rResult.m[_M41] = 0.0f;
    rResult.m[_M42] = 0.0f;
    rResult.m[_M43] = (nearPlaneDistance * farPlaneDistance) / (nearPlaneDistance - farPlaneDistance);
    rResult.m[_M44] = 0.0f;
  }


  Matrix Matrix::CreatePerspectiveFieldOfView(const float fieldOfView, const float aspectRatio, const float nearPlaneDistance, const float farPlaneDistance)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreatePerspectiveFieldOfView(result, fieldOfView, aspectRatio, nearPlaneDistance, farPlaneDistance);
    return result;
  }


  void Matrix::CreatePerspectiveFieldOfView(Matrix& rResult, const float fieldOfView, const float aspectRatio, const float nearPlaneDistance, const float farPlaneDistance)
  {
    if ((fieldOfView <= 0.0f) || (fieldOfView >= 3.141593f))
      throw std::invalid_argument("fieldOfView <= 0 or >= PI");
    if (nearPlaneDistance <= 0.0f)
      throw std::invalid_argument("nearPlaneDistance <= 0");
    if (farPlaneDistance <= 0.0f)
      throw std::invalid_argument("farPlaneDistance <= 0");
    if (nearPlaneDistance >= farPlaneDistance)
      throw std::invalid_argument("nearPlaneDistance >= farPlaneDistance");

    const float num = 1.0f / ((float)std::tan(fieldOfView * 0.5));
    const float num9 = num / aspectRatio;
    rResult.m[_M11] = num9;
    rResult.m[_M12] = 0.0f;
    rResult.m[_M13] = 0.0f;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = 0.0f;
    rResult.m[_M22] = num;
    rResult.m[_M23] = 0.0f;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = 0.0f;
    rResult.m[_M32] = 0.0f;
    rResult.m[_M33] = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
    rResult.m[_M34] = -1.0;
    rResult.m[_M41] = 0.0f;
    rResult.m[_M42] = 0.0f;
    rResult.m[_M43] = (nearPlaneDistance * farPlaneDistance) / (nearPlaneDistance - farPlaneDistance);
    rResult.m[_M44] = 0.0f;
  }

  //Matrix Matrix::CreateReflection(Plane value)
  //{
  //  Matrix result(OptimizationFlag::NoInitialization);
  //  CreateReflection(ref value, out result);
  //  return result;
  //}

  //void Matrix::CreateReflection(Matrix& rResult, const Plane& value)
  //{
  //  Plane plane(OptimizationFlag::NoInitialization);
  //  Plane::Normalize(plane, value);
  //  value.Normalize();
  //  const float x = plane.Normal.X;
  //  const float y = plane.Normal.Y;
  //  const float z = plane.Normal.Z;
  //  const float num3 = -2.0f * x;
  //  const float num2 = -2.0f * y;
  //  const float num = -2.0f * z;
  //  rResult.m[_M11] = (num3 * x) + 1f;
  //  rResult.m[_M12] = num2 * x;
  //  rResult.m[_M13] = num * x;
  //  rResult.m[_M14] = 0.0f;
  //  rResult.m[_M21] = num3 * y;
  //  rResult.m[_M22] = (num2 * y) + 1;
  //  rResult.m[_M23] = num * y;
  //  rResult.m[_M24] = 0.0f;
  //  rResult.m[_M31] = num3 * z;
  //  rResult.m[_M32] = num2 * z;
  //  rResult.m[_M33] = (num * z) + 1;
  //  rResult.m[_M34] = 0.0f;
  //  rResult.m[_M41] = num3 * plane.D;
  //  rResult.m[_M42] = num2 * plane.D;
  //  rResult.m[_M43] = num * plane.D;
  //  rResult.m[_M44] = 1.0f;
  //}


  Matrix Matrix::CreateRotationX(const float radians)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateRotationX(result, radians);
    return result;
  }


  void Matrix::CreateRotationX(Matrix& rResult, const float radians)
  {
    rResult.SetIdentity();

    const float val1 = std::cos(radians);
    const float val2 = std::sin(radians);

    rResult.m[_M22] = val1;
    rResult.m[_M23] = val2;
    rResult.m[_M32] = -val2;
    rResult.m[_M33] = val1;
  }


  Matrix Matrix::CreateRotationY(const float radians)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateRotationY(result, radians);
    return result;
  }


  void Matrix::CreateRotationY(Matrix& rResult, const float radians)
  {
    rResult.SetIdentity();

    const float val1 = std::cos(radians);
    const float val2 = std::sin(radians);

    rResult.m[_M11] = val1;
    rResult.m[_M13] = -val2;
    rResult.m[_M31] = val2;
    rResult.m[_M33] = val1;
  }


  Matrix Matrix::CreateRotationZ(const float radians)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateRotationZ(result, radians);
    return result;
  }


  void Matrix::CreateRotationZ(Matrix& rResult, const float radians)
  {
    rResult.SetIdentity();

    const float val1 = std::cos(radians);
    const float val2 = std::sin(radians);

    rResult.m[_M11] = val1;
    rResult.m[_M12] = val2;
    rResult.m[_M21] = -val2;
    rResult.m[_M22] = val1;
  }


  Matrix Matrix::CreateScale(const float scale)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateScale(result, scale, scale, scale);
    return result;
  }

  void Matrix::CreateScale(Matrix& rResult, const float scale)
  {
    CreateScale(rResult, scale, scale, scale);
  }

  Matrix Matrix::CreateScale(const float scaleX, const float scaleY, const float scaleZ)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateScale(result, scaleX, scaleY, scaleZ);
    return result;
  }

  void Matrix::CreateScale(Matrix& rResult, const float scaleX, const float scaleY, const float scaleZ)
  {
    rResult.m[_M11] = scaleX;
    rResult.m[_M12] = 0;
    rResult.m[_M13] = 0;
    rResult.m[_M14] = 0;
    rResult.m[_M21] = 0;
    rResult.m[_M22] = scaleY;
    rResult.m[_M23] = 0;
    rResult.m[_M24] = 0;
    rResult.m[_M31] = 0;
    rResult.m[_M32] = 0;
    rResult.m[_M33] = scaleZ;
    rResult.m[_M34] = 0;
    rResult.m[_M41] = 0;
    rResult.m[_M42] = 0;
    rResult.m[_M43] = 0;
    rResult.m[_M44] = 1;
  }

  Matrix Matrix::CreateScale(const Vector3& scale)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateScale(result, scale);
    return result;
  }

  void Matrix::CreateScale(Matrix& rResult, const Vector3& scale)
  {
    rResult.m[_M11] = scale.X;
    rResult.m[_M12] = 0;
    rResult.m[_M13] = 0;
    rResult.m[_M14] = 0;
    rResult.m[_M21] = 0;
    rResult.m[_M22] = scale.Y;
    rResult.m[_M23] = 0;
    rResult.m[_M24] = 0;
    rResult.m[_M31] = 0;
    rResult.m[_M32] = 0;
    rResult.m[_M33] = scale.Z;
    rResult.m[_M34] = 0;
    rResult.m[_M41] = 0;
    rResult.m[_M42] = 0;
    rResult.m[_M43] = 0;
    rResult.m[_M44] = 1;
  }


  //Matrix Matrix::CreateShadow(const Vector3& lightDirection, const Plane& plane)
  //{
  //  Matrix result(OptimizationFlag::NoInitialization);
  //  CreateShadow(result, lightDirection, plane);
  //  return result;
  //}


  //void Matrix::CreateShadow(Matrix& rResult, const Vector3& lightDirection, const Plane& plane)
  //{
  //  const float dot = (plane.Normal.X * lightDirection.X) + (plane.Normal.Y * lightDirection.Y) + (plane.Normal.Z * lightDirection.Z);
  //  const float x = -plane.Normal.X;
  //  const float y = -plane.Normal.Y;
  //  const float z = -plane.Normal.Z;
  //  const float d = -plane.D;

  //  rResult.m[_M11] = (x * lightDirection.X) + dot;
  //  rResult.m[_M12] = x * lightDirection.Y;
  //  rResult.m[_M13] = x * lightDirection.Z;
  //  rResult.m[_M14] = 0.0f;
  //  rResult.m[_M21] = y * lightDirection.X;
  //  rResult.m[_M22] = (y * lightDirection.Y) + dot;
  //  rResult.m[_M23] = y * lightDirection.Z;
  //  rResult.m[_M24] = 0.0f;
  //  rResult.m[_M31] = z * lightDirection.X;
  //  rResult.m[_M32] = z * lightDirection.Y;
  //  rResult.m[_M33] = (z * lightDirection.Z) + dot;
  //  rResult.m[_M34] = 0.0f;
  //  rResult.m[_M41] = d * lightDirection.X;
  //  rResult.m[_M42] = d * lightDirection.Y;
  //  rResult.m[_M43] = d * lightDirection.Z;
  //  rResult.m[_M44] = dot;
  //}


  Matrix Matrix::CreateTranslation(const float x, const float y, const float z)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateTranslation(result, x, y, z);
    return result;
  }


  void Matrix::CreateTranslation(Matrix& rResult, const float x, const float y, const float z)
  {
    rResult.m[_M11] = 1.0f;
    rResult.m[_M12] = 0.0f;
    rResult.m[_M13] = 0.0f;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = 0.0f;
    rResult.m[_M22] = 1.0f;
    rResult.m[_M23] = 0.0f;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = 0.0f;
    rResult.m[_M32] = 0.0f;
    rResult.m[_M33] = 1.0f;
    rResult.m[_M34] = 0.0f;
    rResult.m[_M41] = x;
    rResult.m[_M42] = y;
    rResult.m[_M43] = z;
    rResult.m[_M44] = 1.0f;
  }


  Matrix Matrix::CreateTranslation(const Vector3& value)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateTranslation(result, value);
    return result;
  }


  void Matrix::CreateTranslation(Matrix& rResult, const Vector3& value)
  {
    rResult.m[_M11] = 1.0f;
    rResult.m[_M12] = 0.0f;
    rResult.m[_M13] = 0.0f;
    rResult.m[_M14] = 0.0f;
    rResult.m[_M21] = 0.0f;
    rResult.m[_M22] = 1.0f;
    rResult.m[_M23] = 0.0f;
    rResult.m[_M24] = 0.0f;
    rResult.m[_M31] = 0.0f;
    rResult.m[_M32] = 0.0f;
    rResult.m[_M33] = 1.0f;
    rResult.m[_M34] = 0.0f;
    rResult.m[_M41] = value.X;
    rResult.m[_M42] = value.Y;
    rResult.m[_M43] = value.Z;
    rResult.m[_M44] = 1.0f;
  }


  Matrix Matrix::CreateWorld(const Vector3& position, const Vector3& forward, const Vector3& up)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    CreateWorld(result, position, forward, up);
    return result;
  }


  void Matrix::CreateWorld(Matrix& rResult, const Vector3& position, const Vector3& forward, const Vector3& up)
  {
    Vector3 x, y, z;
    Vector3::Normalize(z, forward);
    Vector3::Cross(x, forward, up);
    Vector3::Cross(y, x, forward);
    x.Normalize();
    y.Normalize();

    rResult.Clear();
    rResult.SetRight(x);
    rResult.SetUp(y);
    rResult.SetForward(z);
    rResult.SetTranslation(position);
    rResult.m[_M44] = 1.0f;
  }


  float Matrix::Determinant() const
  {
    const float num22 = m[_M11];
    const float num21 = m[_M12];
    const float num20 = m[_M13];
    const float num19 = m[_M14];
    const float num12 = m[_M21];
    const float num11 = m[_M22];
    const float num10 = m[_M23];
    const float num9 = m[_M24];
    const float num8 = m[_M31];
    const float num7 = m[_M32];
    const float num6 = m[_M33];
    const float num5 = m[_M34];
    const float num4 = m[_M41];
    const float num3 = m[_M42];
    const float num2 = m[_M43];
    const float num = m[_M44];
    const float num18 = (num6 * num) - (num5 * num2);
    const float num17 = (num7 * num) - (num5 * num3);
    const float num16 = (num7 * num2) - (num6 * num3);
    const float num15 = (num8 * num) - (num5 * num4);
    const float num14 = (num8 * num2) - (num6 * num4);
    const float num13 = (num8 * num3) - (num7 * num4);
    return ((((num22 * (((num11 * num18) - (num10 * num17)) + (num9 * num16))) - (num21 * (((num12 * num18) - (num10 * num15)) + (num9 * num14)))) + (num20 * (((num12 * num17) - (num11 * num15)) + (num9 * num13)))) - (num19 * (((num12 * num16) - (num11 * num14)) + (num10 * num13))));
  }


  Matrix Matrix::Divide(const Matrix& matrix1, const Matrix& matrix2)
  {
    return Matrix(
      matrix1.m[_M11] / matrix2.m[_M11],
      matrix1.m[_M12] / matrix2.m[_M12],
      matrix1.m[_M13] / matrix2.m[_M13],
      matrix1.m[_M14] / matrix2.m[_M14],
      matrix1.m[_M21] / matrix2.m[_M21],
      matrix1.m[_M22] / matrix2.m[_M22],
      matrix1.m[_M23] / matrix2.m[_M23],
      matrix1.m[_M24] / matrix2.m[_M24],
      matrix1.m[_M31] / matrix2.m[_M31],
      matrix1.m[_M32] / matrix2.m[_M32],
      matrix1.m[_M33] / matrix2.m[_M33],
      matrix1.m[_M34] / matrix2.m[_M34],
      matrix1.m[_M41] / matrix2.m[_M41],
      matrix1.m[_M42] / matrix2.m[_M42],
      matrix1.m[_M43] / matrix2.m[_M43],
      matrix1.m[_M44] / matrix2.m[_M44]);
  }


  void Matrix::Divide(Matrix& rResult, const Matrix& matrix1, const Matrix& matrix2)
  {
    rResult.m[_M11] = matrix1.m[_M11] / matrix2.m[_M11];
    rResult.m[_M12] = matrix1.m[_M12] / matrix2.m[_M12];
    rResult.m[_M13] = matrix1.m[_M13] / matrix2.m[_M13];
    rResult.m[_M14] = matrix1.m[_M14] / matrix2.m[_M14];
    rResult.m[_M21] = matrix1.m[_M21] / matrix2.m[_M21];
    rResult.m[_M22] = matrix1.m[_M22] / matrix2.m[_M22];
    rResult.m[_M23] = matrix1.m[_M23] / matrix2.m[_M23];
    rResult.m[_M24] = matrix1.m[_M24] / matrix2.m[_M24];
    rResult.m[_M31] = matrix1.m[_M31] / matrix2.m[_M31];
    rResult.m[_M32] = matrix1.m[_M32] / matrix2.m[_M32];
    rResult.m[_M33] = matrix1.m[_M33] / matrix2.m[_M33];
    rResult.m[_M34] = matrix1.m[_M34] / matrix2.m[_M34];
    rResult.m[_M41] = matrix1.m[_M41] / matrix2.m[_M41];
    rResult.m[_M42] = matrix1.m[_M42] / matrix2.m[_M42];
    rResult.m[_M43] = matrix1.m[_M43] / matrix2.m[_M43];
    rResult.m[_M44] = matrix1.m[_M44] / matrix2.m[_M44];
  }


  Matrix Matrix::Divide(const Matrix& matrix1, const float divider)
  {
    // Multiply usually has better performance
    const float num = 1.0f / divider;

    return Matrix(
      matrix1.m[_M11] * num, matrix1.m[_M12] * num, matrix1.m[_M13] * num, matrix1.m[_M14] * num,
      matrix1.m[_M21] * num, matrix1.m[_M22] * num, matrix1.m[_M23] * num, matrix1.m[_M24] * num,
      matrix1.m[_M31] * num, matrix1.m[_M32] * num, matrix1.m[_M33] * num, matrix1.m[_M34] * num,
      matrix1.m[_M41] * num, matrix1.m[_M42] * num, matrix1.m[_M43] * num, matrix1.m[_M44] * num);
  }


  void Matrix::Divide(Matrix& rResult, const Matrix& matrix1, const float divider)
  {
    // Multiply usually has better performance
    const float num = 1.0f / divider;
    rResult.m[_M11] = matrix1.m[_M11] * num;
    rResult.m[_M12] = matrix1.m[_M12] * num;
    rResult.m[_M13] = matrix1.m[_M13] * num;
    rResult.m[_M14] = matrix1.m[_M14] * num;
    rResult.m[_M21] = matrix1.m[_M21] * num;
    rResult.m[_M22] = matrix1.m[_M22] * num;
    rResult.m[_M23] = matrix1.m[_M23] * num;
    rResult.m[_M24] = matrix1.m[_M24] * num;
    rResult.m[_M31] = matrix1.m[_M31] * num;
    rResult.m[_M32] = matrix1.m[_M32] * num;
    rResult.m[_M33] = matrix1.m[_M33] * num;
    rResult.m[_M34] = matrix1.m[_M34] * num;
    rResult.m[_M41] = matrix1.m[_M41] * num;
    rResult.m[_M42] = matrix1.m[_M42] * num;
    rResult.m[_M43] = matrix1.m[_M43] * num;
    rResult.m[_M44] = matrix1.m[_M44] * num;
  }


  Matrix Matrix::Lerp(const Matrix& value1, const Matrix& value2, const float amount)
  {
    const float* pMatrix1 = value1.m;
    const float* pMatrix2 = value2.m;
    return Matrix(
      pMatrix1[_M11] + ((pMatrix2[_M11] - pMatrix1[_M11]) * amount),
      pMatrix1[_M12] + ((pMatrix2[_M12] - pMatrix1[_M12]) * amount),
      pMatrix1[_M13] + ((pMatrix2[_M13] - pMatrix1[_M13]) * amount),
      pMatrix1[_M14] + ((pMatrix2[_M14] - pMatrix1[_M14]) * amount),
      pMatrix1[_M21] + ((pMatrix2[_M21] - pMatrix1[_M21]) * amount),
      pMatrix1[_M22] + ((pMatrix2[_M22] - pMatrix1[_M22]) * amount),
      pMatrix1[_M23] + ((pMatrix2[_M23] - pMatrix1[_M23]) * amount),
      pMatrix1[_M24] + ((pMatrix2[_M24] - pMatrix1[_M24]) * amount),
      pMatrix1[_M31] + ((pMatrix2[_M31] - pMatrix1[_M31]) * amount),
      pMatrix1[_M32] + ((pMatrix2[_M32] - pMatrix1[_M32]) * amount),
      pMatrix1[_M33] + ((pMatrix2[_M33] - pMatrix1[_M33]) * amount),
      pMatrix1[_M34] + ((pMatrix2[_M34] - pMatrix1[_M34]) * amount),
      pMatrix1[_M41] + ((pMatrix2[_M41] - pMatrix1[_M41]) * amount),
      pMatrix1[_M42] + ((pMatrix2[_M42] - pMatrix1[_M42]) * amount),
      pMatrix1[_M43] + ((pMatrix2[_M43] - pMatrix1[_M43]) * amount),
      pMatrix1[_M44] + ((pMatrix2[_M44] - pMatrix1[_M44]) * amount));
  }


  void Matrix::Lerp(Matrix& rResult, const Matrix& value1, const Matrix& value2, const float amount)
  {
    const float* pMatrix1 = value1.m;
    const float* pMatrix2 = value2.m;
    float* pResult = rResult.m;

    // It's not recommended to write the result to the matrixes that are being used to interpolate between
    assert(pMatrix1 != pResult);
    assert(pMatrix2 != pResult);

    pResult[_M11] = pMatrix1[_M11] + ((pMatrix2[_M11] - pMatrix1[_M11]) * amount);
    pResult[_M12] = pMatrix1[_M12] + ((pMatrix2[_M12] - pMatrix1[_M12]) * amount);
    pResult[_M13] = pMatrix1[_M13] + ((pMatrix2[_M13] - pMatrix1[_M13]) * amount);
    pResult[_M14] = pMatrix1[_M14] + ((pMatrix2[_M14] - pMatrix1[_M14]) * amount);
    pResult[_M21] = pMatrix1[_M21] + ((pMatrix2[_M21] - pMatrix1[_M21]) * amount);
    pResult[_M22] = pMatrix1[_M22] + ((pMatrix2[_M22] - pMatrix1[_M22]) * amount);
    pResult[_M23] = pMatrix1[_M23] + ((pMatrix2[_M23] - pMatrix1[_M23]) * amount);
    pResult[_M24] = pMatrix1[_M24] + ((pMatrix2[_M24] - pMatrix1[_M24]) * amount);
    pResult[_M31] = pMatrix1[_M31] + ((pMatrix2[_M31] - pMatrix1[_M31]) * amount);
    pResult[_M32] = pMatrix1[_M32] + ((pMatrix2[_M32] - pMatrix1[_M32]) * amount);
    pResult[_M33] = pMatrix1[_M33] + ((pMatrix2[_M33] - pMatrix1[_M33]) * amount);
    pResult[_M34] = pMatrix1[_M34] + ((pMatrix2[_M34] - pMatrix1[_M34]) * amount);
    pResult[_M41] = pMatrix1[_M41] + ((pMatrix2[_M41] - pMatrix1[_M41]) * amount);
    pResult[_M42] = pMatrix1[_M42] + ((pMatrix2[_M42] - pMatrix1[_M42]) * amount);
    pResult[_M43] = pMatrix1[_M43] + ((pMatrix2[_M43] - pMatrix1[_M43]) * amount);
    pResult[_M44] = pMatrix1[_M44] + ((pMatrix2[_M44] - pMatrix1[_M44]) * amount);
  }

  Matrix Matrix::Multiply(const Matrix& matrix1, const Matrix& matrix2)
  {
    return Matrix(
    (((matrix1.m[_M11] * matrix2.m[_M11]) + (matrix1.m[_M12] * matrix2.m[_M21])) + (matrix1.m[_M13] * matrix2.m[_M31])) + (matrix1.m[_M14] * matrix2.m[_M41]),
    (((matrix1.m[_M11] * matrix2.m[_M12]) + (matrix1.m[_M12] * matrix2.m[_M22])) + (matrix1.m[_M13] * matrix2.m[_M32])) + (matrix1.m[_M14] * matrix2.m[_M42]),
    (((matrix1.m[_M11] * matrix2.m[_M13]) + (matrix1.m[_M12] * matrix2.m[_M23])) + (matrix1.m[_M13] * matrix2.m[_M33])) + (matrix1.m[_M14] * matrix2.m[_M43]),
    (((matrix1.m[_M11] * matrix2.m[_M14]) + (matrix1.m[_M12] * matrix2.m[_M24])) + (matrix1.m[_M13] * matrix2.m[_M34])) + (matrix1.m[_M14] * matrix2.m[_M44]),
    (((matrix1.m[_M21] * matrix2.m[_M11]) + (matrix1.m[_M22] * matrix2.m[_M21])) + (matrix1.m[_M23] * matrix2.m[_M31])) + (matrix1.m[_M24] * matrix2.m[_M41]),
    (((matrix1.m[_M21] * matrix2.m[_M12]) + (matrix1.m[_M22] * matrix2.m[_M22])) + (matrix1.m[_M23] * matrix2.m[_M32])) + (matrix1.m[_M24] * matrix2.m[_M42]),
    (((matrix1.m[_M21] * matrix2.m[_M13]) + (matrix1.m[_M22] * matrix2.m[_M23])) + (matrix1.m[_M23] * matrix2.m[_M33])) + (matrix1.m[_M24] * matrix2.m[_M43]),
    (((matrix1.m[_M21] * matrix2.m[_M14]) + (matrix1.m[_M22] * matrix2.m[_M24])) + (matrix1.m[_M23] * matrix2.m[_M34])) + (matrix1.m[_M24] * matrix2.m[_M44]),
    (((matrix1.m[_M31] * matrix2.m[_M11]) + (matrix1.m[_M32] * matrix2.m[_M21])) + (matrix1.m[_M33] * matrix2.m[_M31])) + (matrix1.m[_M34] * matrix2.m[_M41]),
    (((matrix1.m[_M31] * matrix2.m[_M12]) + (matrix1.m[_M32] * matrix2.m[_M22])) + (matrix1.m[_M33] * matrix2.m[_M32])) + (matrix1.m[_M34] * matrix2.m[_M42]),
    (((matrix1.m[_M31] * matrix2.m[_M13]) + (matrix1.m[_M32] * matrix2.m[_M23])) + (matrix1.m[_M33] * matrix2.m[_M33])) + (matrix1.m[_M34] * matrix2.m[_M43]),
    (((matrix1.m[_M31] * matrix2.m[_M14]) + (matrix1.m[_M32] * matrix2.m[_M24])) + (matrix1.m[_M33] * matrix2.m[_M34])) + (matrix1.m[_M34] * matrix2.m[_M44]),
    (((matrix1.m[_M41] * matrix2.m[_M11]) + (matrix1.m[_M42] * matrix2.m[_M21])) + (matrix1.m[_M43] * matrix2.m[_M31])) + (matrix1.m[_M44] * matrix2.m[_M41]),
    (((matrix1.m[_M41] * matrix2.m[_M12]) + (matrix1.m[_M42] * matrix2.m[_M22])) + (matrix1.m[_M43] * matrix2.m[_M32])) + (matrix1.m[_M44] * matrix2.m[_M42]),
    (((matrix1.m[_M41] * matrix2.m[_M13]) + (matrix1.m[_M42] * matrix2.m[_M23])) + (matrix1.m[_M43] * matrix2.m[_M33])) + (matrix1.m[_M44] * matrix2.m[_M43]),
    (((matrix1.m[_M41] * matrix2.m[_M14]) + (matrix1.m[_M42] * matrix2.m[_M24])) + (matrix1.m[_M43] * matrix2.m[_M34])) + (matrix1.m[_M44] * matrix2.m[_M44]));
  }


  void Matrix::Multiply(Matrix& rResult, const Matrix& matrix1, const Matrix& matrix2)
  {
    const float* pMatrix1 = matrix1.m;
    const float* pMatrix2 = matrix2.m;

    const float m11 = (((pMatrix1[_M11] * pMatrix2[_M11]) + (pMatrix1[_M12] * pMatrix2[_M21])) + (pMatrix1[_M13] * pMatrix2[_M31])) + (pMatrix1[_M14] * pMatrix2[_M41]);
    const float m12 = (((pMatrix1[_M11] * pMatrix2[_M12]) + (pMatrix1[_M12] * pMatrix2[_M22])) + (pMatrix1[_M13] * pMatrix2[_M32])) + (pMatrix1[_M14] * pMatrix2[_M42]);
    const float m13 = (((pMatrix1[_M11] * pMatrix2[_M13]) + (pMatrix1[_M12] * pMatrix2[_M23])) + (pMatrix1[_M13] * pMatrix2[_M33])) + (pMatrix1[_M14] * pMatrix2[_M43]);
    const float m14 = (((pMatrix1[_M11] * pMatrix2[_M14]) + (pMatrix1[_M12] * pMatrix2[_M24])) + (pMatrix1[_M13] * pMatrix2[_M34])) + (pMatrix1[_M14] * pMatrix2[_M44]);
    const float m21 = (((pMatrix1[_M21] * pMatrix2[_M11]) + (pMatrix1[_M22] * pMatrix2[_M21])) + (pMatrix1[_M23] * pMatrix2[_M31])) + (pMatrix1[_M24] * pMatrix2[_M41]);
    const float m22 = (((pMatrix1[_M21] * pMatrix2[_M12]) + (pMatrix1[_M22] * pMatrix2[_M22])) + (pMatrix1[_M23] * pMatrix2[_M32])) + (pMatrix1[_M24] * pMatrix2[_M42]);
    const float m23 = (((pMatrix1[_M21] * pMatrix2[_M13]) + (pMatrix1[_M22] * pMatrix2[_M23])) + (pMatrix1[_M23] * pMatrix2[_M33])) + (pMatrix1[_M24] * pMatrix2[_M43]);
    const float m24 = (((pMatrix1[_M21] * pMatrix2[_M14]) + (pMatrix1[_M22] * pMatrix2[_M24])) + (pMatrix1[_M23] * pMatrix2[_M34])) + (pMatrix1[_M24] * pMatrix2[_M44]);
    const float m31 = (((pMatrix1[_M31] * pMatrix2[_M11]) + (pMatrix1[_M32] * pMatrix2[_M21])) + (pMatrix1[_M33] * pMatrix2[_M31])) + (pMatrix1[_M34] * pMatrix2[_M41]);
    const float m32 = (((pMatrix1[_M31] * pMatrix2[_M12]) + (pMatrix1[_M32] * pMatrix2[_M22])) + (pMatrix1[_M33] * pMatrix2[_M32])) + (pMatrix1[_M34] * pMatrix2[_M42]);
    const float m33 = (((pMatrix1[_M31] * pMatrix2[_M13]) + (pMatrix1[_M32] * pMatrix2[_M23])) + (pMatrix1[_M33] * pMatrix2[_M33])) + (pMatrix1[_M34] * pMatrix2[_M43]);
    const float m34 = (((pMatrix1[_M31] * pMatrix2[_M14]) + (pMatrix1[_M32] * pMatrix2[_M24])) + (pMatrix1[_M33] * pMatrix2[_M34])) + (pMatrix1[_M34] * pMatrix2[_M44]);
    const float m41 = (((pMatrix1[_M41] * pMatrix2[_M11]) + (pMatrix1[_M42] * pMatrix2[_M21])) + (pMatrix1[_M43] * pMatrix2[_M31])) + (pMatrix1[_M44] * pMatrix2[_M41]);
    const float m42 = (((pMatrix1[_M41] * pMatrix2[_M12]) + (pMatrix1[_M42] * pMatrix2[_M22])) + (pMatrix1[_M43] * pMatrix2[_M32])) + (pMatrix1[_M44] * pMatrix2[_M42]);
    const float m43 = (((pMatrix1[_M41] * pMatrix2[_M13]) + (pMatrix1[_M42] * pMatrix2[_M23])) + (pMatrix1[_M43] * pMatrix2[_M33])) + (pMatrix1[_M44] * pMatrix2[_M43]);
    const float m44 = (((pMatrix1[_M41] * pMatrix2[_M14]) + (pMatrix1[_M42] * pMatrix2[_M24])) + (pMatrix1[_M43] * pMatrix2[_M34])) + (pMatrix1[_M44] * pMatrix2[_M44]);

    rResult.m[_M11] = m11;
    rResult.m[_M12] = m12;
    rResult.m[_M13] = m13;
    rResult.m[_M14] = m14;
    rResult.m[_M21] = m21;
    rResult.m[_M22] = m22;
    rResult.m[_M23] = m23;
    rResult.m[_M24] = m24;
    rResult.m[_M31] = m31;
    rResult.m[_M32] = m32;
    rResult.m[_M33] = m33;
    rResult.m[_M34] = m34;
    rResult.m[_M41] = m41;
    rResult.m[_M42] = m42;
    rResult.m[_M43] = m43;
    rResult.m[_M44] = m44;
  }


  Matrix Matrix::Multiply(const Matrix& matrix1, const float factor)
  {
    return Matrix(
      matrix1.m[_M11] * factor, matrix1.m[_M12] * factor, matrix1.m[_M13] * factor, matrix1.m[_M14] * factor,
      matrix1.m[_M21] * factor, matrix1.m[_M22] * factor, matrix1.m[_M23] * factor, matrix1.m[_M24] * factor,
      matrix1.m[_M31] * factor, matrix1.m[_M32] * factor, matrix1.m[_M33] * factor, matrix1.m[_M34] * factor,
      matrix1.m[_M41] * factor, matrix1.m[_M42] * factor, matrix1.m[_M43] * factor, matrix1.m[_M44] * factor);
  }

  void Matrix::Multiply(Matrix& rResult, const Matrix& matrix1, const float factor)
  {
    rResult.m[_M11] = matrix1.m[_M11] * factor;
    rResult.m[_M12] = matrix1.m[_M12] * factor;
    rResult.m[_M13] = matrix1.m[_M13] * factor;
    rResult.m[_M14] = matrix1.m[_M14] * factor;
    rResult.m[_M21] = matrix1.m[_M21] * factor;
    rResult.m[_M22] = matrix1.m[_M22] * factor;
    rResult.m[_M23] = matrix1.m[_M23] * factor;
    rResult.m[_M24] = matrix1.m[_M24] * factor;
    rResult.m[_M31] = matrix1.m[_M31] * factor;
    rResult.m[_M32] = matrix1.m[_M32] * factor;
    rResult.m[_M33] = matrix1.m[_M33] * factor;
    rResult.m[_M34] = matrix1.m[_M34] * factor;
    rResult.m[_M41] = matrix1.m[_M41] * factor;
    rResult.m[_M42] = matrix1.m[_M42] * factor;
    rResult.m[_M43] = matrix1.m[_M43] * factor;
    rResult.m[_M44] = matrix1.m[_M44] * factor;
  }


  Matrix Matrix::Negate(const Matrix& matrix)
  {
    return Matrix(
      -matrix.m[_M11], -matrix.m[_M12], -matrix.m[_M13], -matrix.m[_M14],
      -matrix.m[_M21], -matrix.m[_M22], -matrix.m[_M23], -matrix.m[_M24],
      -matrix.m[_M31], -matrix.m[_M32], -matrix.m[_M33], -matrix.m[_M34],
      -matrix.m[_M41], -matrix.m[_M42], -matrix.m[_M43], -matrix.m[_M44]
      );
  }


  void Matrix::Negate(Matrix& rResult, const Matrix& matrix)
  {
    rResult.m[_M11] = -matrix.m[_M11];
    rResult.m[_M12] = -matrix.m[_M12];
    rResult.m[_M13] = -matrix.m[_M13];
    rResult.m[_M14] = -matrix.m[_M14];
    rResult.m[_M21] = -matrix.m[_M21];
    rResult.m[_M22] = -matrix.m[_M22];
    rResult.m[_M23] = -matrix.m[_M23];
    rResult.m[_M24] = -matrix.m[_M24];
    rResult.m[_M31] = -matrix.m[_M31];
    rResult.m[_M32] = -matrix.m[_M32];
    rResult.m[_M33] = -matrix.m[_M33];
    rResult.m[_M34] = -matrix.m[_M34];
    rResult.m[_M41] = -matrix.m[_M41];
    rResult.m[_M42] = -matrix.m[_M42];
    rResult.m[_M43] = -matrix.m[_M43];
    rResult.m[_M44] = -matrix.m[_M44];
  }


  void Matrix::SetIdentity()
  {
    m[_M11] = 1.0f; m[_M12] = 0.0f; m[_M13] = 0.0f; m[_M14] = 0.0f;
    m[_M21] = 0.0f; m[_M22] = 1.0f; m[_M23] = 0.0f; m[_M24] = 0.0f;
    m[_M31] = 0.0f; m[_M32] = 0.0f; m[_M33] = 1.0f; m[_M34] = 0.0f;
    m[_M41] = 0.0f; m[_M42] = 0.0f; m[_M43] = 0.0f; m[_M44] = 1.0f;
  }

  Matrix Matrix::Subtract(const Matrix& matrix1, const Matrix& matrix2)
  {
    return Matrix(
      matrix1.m[_M11] - matrix2.m[_M11],
      matrix1.m[_M12] - matrix2.m[_M12],
      matrix1.m[_M13] - matrix2.m[_M13],
      matrix1.m[_M14] - matrix2.m[_M14],
      matrix1.m[_M21] - matrix2.m[_M21],
      matrix1.m[_M22] - matrix2.m[_M22],
      matrix1.m[_M23] - matrix2.m[_M23],
      matrix1.m[_M24] - matrix2.m[_M24],
      matrix1.m[_M31] - matrix2.m[_M31],
      matrix1.m[_M32] - matrix2.m[_M32],
      matrix1.m[_M33] - matrix2.m[_M33],
      matrix1.m[_M34] - matrix2.m[_M34],
      matrix1.m[_M41] - matrix2.m[_M41],
      matrix1.m[_M42] - matrix2.m[_M42],
      matrix1.m[_M43] - matrix2.m[_M43],
      matrix1.m[_M44] - matrix2.m[_M44]);
  }


  void Matrix::Subtract(Matrix& rResult, const Matrix& matrix1, const Matrix& matrix2)
  {
    rResult.m[_M11] = matrix1.m[_M11] - matrix2.m[_M11];
    rResult.m[_M12] = matrix1.m[_M12] - matrix2.m[_M12];
    rResult.m[_M13] = matrix1.m[_M13] - matrix2.m[_M13];
    rResult.m[_M14] = matrix1.m[_M14] - matrix2.m[_M14];
    rResult.m[_M21] = matrix1.m[_M21] - matrix2.m[_M21];
    rResult.m[_M22] = matrix1.m[_M22] - matrix2.m[_M22];
    rResult.m[_M23] = matrix1.m[_M23] - matrix2.m[_M23];
    rResult.m[_M24] = matrix1.m[_M24] - matrix2.m[_M24];
    rResult.m[_M31] = matrix1.m[_M31] - matrix2.m[_M31];
    rResult.m[_M32] = matrix1.m[_M32] - matrix2.m[_M32];
    rResult.m[_M33] = matrix1.m[_M33] - matrix2.m[_M33];
    rResult.m[_M34] = matrix1.m[_M34] - matrix2.m[_M34];
    rResult.m[_M41] = matrix1.m[_M41] - matrix2.m[_M41];
    rResult.m[_M42] = matrix1.m[_M42] - matrix2.m[_M42];
    rResult.m[_M43] = matrix1.m[_M43] - matrix2.m[_M43];
    rResult.m[_M44] = matrix1.m[_M44] - matrix2.m[_M44];
  }


  Matrix Matrix::Transpose(const Matrix& matrix)
  {
    Matrix result(OptimizationFlag::NoInitialization);
    Transpose(result, matrix);
    return result;
  }


  void Matrix::Transpose(Matrix& rResult, const Matrix& src)
  {
    // Copy to the same position
    rResult.m[_M11] = src.m[_M11];
    rResult.m[_M22] = src.m[_M22];
    rResult.m[_M33] = src.m[_M33];
    rResult.m[_M44] = src.m[_M44];

    // We use the tmp variable to allow in place transpose
    float tmp;
    tmp = src.m[_M12];
    rResult.m[_M12] = src.m[_M21];
    rResult.m[_M21] = tmp;

    tmp = src.m[_M13];
    rResult.m[_M13] = src.m[_M31];
    rResult.m[_M31] = tmp;

    tmp = src.m[_M14];
    rResult.m[_M14] = src.m[_M41];
    rResult.m[_M41] = tmp;

    tmp = src.m[_M23];
    rResult.m[_M23] = src.m[_M32];
    rResult.m[_M32] = tmp;

    tmp = src.m[_M24];
    rResult.m[_M24] = src.m[_M42];
    rResult.m[_M42] = tmp;

    tmp = src.m[_M34];
    rResult.m[_M34] = src.m[_M43];
    rResult.m[_M43] = tmp;
  }


  void MatrixInternals::Transform(Vector2& rResult, const Vector2& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector2((position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M21]) + pMatrix[_M41],
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M22]) + pMatrix[_M42]);
  }


  void MatrixInternals::TransformNormal(Vector2& rResult, const Vector2& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector2((position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M21]),
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M22]));
  }

  void MatrixInternals::Transform(Vector3& rResult, const Vector3& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector3((position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M21]) + (position.Z * pMatrix[_M31]) + pMatrix[_M41],
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M22]) + (position.Z * pMatrix[_M32]) + pMatrix[_M42],
                      (position.X * pMatrix[_M14]) + (position.Y * pMatrix[_M23]) + (position.Z * pMatrix[_M33]) + pMatrix[_M43]);
  }


  void MatrixInternals::TransformNormal(Vector3& rResult, const Vector3& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector3((position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M21]) + (position.Z * pMatrix[_M31]),
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M22]) + (position.Z * pMatrix[_M32]),
                      (position.X * pMatrix[_M14]) + (position.Y * pMatrix[_M23]) + (position.Z * pMatrix[_M33]));
  }

  void MatrixInternals::Transform(Vector4& rResult, const Vector2& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector4((position.X * pMatrix[_M11]) + (position.Y * pMatrix[_M21]) + pMatrix[_M41],
                      (position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M22]) + pMatrix[_M42],
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M23]) + pMatrix[_M43],
                      (position.X * pMatrix[_M14]) + (position.Y * pMatrix[_M24]) + pMatrix[_M44]);
  }

  void MatrixInternals::Transform(Vector4& rResult, const Vector3& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector4((position.X * pMatrix[_M11]) + (position.Y * pMatrix[_M21]) + (position.Z * pMatrix[_M31]) + pMatrix[_M41],
                      (position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M22]) + (position.Z * pMatrix[_M32]) + pMatrix[_M42],
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M23]) + (position.Z * pMatrix[_M33]) + pMatrix[_M43],
                      (position.X * pMatrix[_M14]) + (position.Y * pMatrix[_M24]) + (position.Z * pMatrix[_M34]) + pMatrix[_M44]);
  }


  void MatrixInternals::Transform(Vector4& rResult, const Vector4& position, const Matrix& matrix)
  {
    const float* pMatrix = matrix.DirectAccess();
    rResult = Vector4((position.X * pMatrix[_M11]) + (position.Y * pMatrix[_M21]) + (position.Z * pMatrix[_M31]) + (position.W * pMatrix[_M41]),
                      (position.X * pMatrix[_M12]) + (position.Y * pMatrix[_M22]) + (position.Z * pMatrix[_M32]) + (position.W * pMatrix[_M42]),
                      (position.X * pMatrix[_M13]) + (position.Y * pMatrix[_M23]) + (position.Z * pMatrix[_M33]) + (position.W * pMatrix[_M43]),
                      (position.X * pMatrix[_M14]) + (position.Y * pMatrix[_M24]) + (position.Z * pMatrix[_M34]) + (position.W * pMatrix[_M44]));
  }


  Matrix& Matrix::operator *= (const Matrix& rhs)
  {
    const float* pRhs = rhs.DirectAccess();

    const float m11 = (((m[_M11] * pRhs[_M11]) + (m[_M12] * pRhs[_M21]) + (m[_M13] * pRhs[_M31])) + (m[_M14] * pRhs[_M41]));
    const float m12 = (((m[_M11] * pRhs[_M12]) + (m[_M12] * pRhs[_M22]) + (m[_M13] * pRhs[_M32])) + (m[_M14] * pRhs[_M42]));
    const float m13 = (((m[_M11] * pRhs[_M13]) + (m[_M12] * pRhs[_M23]) + (m[_M13] * pRhs[_M33])) + (m[_M14] * pRhs[_M43]));
    const float m14 = (((m[_M11] * pRhs[_M14]) + (m[_M12] * pRhs[_M24]) + (m[_M13] * pRhs[_M34])) + (m[_M14] * pRhs[_M44]));
    const float m21 = (((m[_M21] * pRhs[_M11]) + (m[_M22] * pRhs[_M21]) + (m[_M23] * pRhs[_M31])) + (m[_M24] * pRhs[_M41]));
    const float m22 = (((m[_M21] * pRhs[_M12]) + (m[_M22] * pRhs[_M22]) + (m[_M23] * pRhs[_M32])) + (m[_M24] * pRhs[_M42]));
    const float m23 = (((m[_M21] * pRhs[_M13]) + (m[_M22] * pRhs[_M23]) + (m[_M23] * pRhs[_M33])) + (m[_M24] * pRhs[_M43]));
    const float m24 = (((m[_M21] * pRhs[_M14]) + (m[_M22] * pRhs[_M24]) + (m[_M23] * pRhs[_M34])) + (m[_M24] * pRhs[_M44]));
    const float m31 = (((m[_M31] * pRhs[_M11]) + (m[_M32] * pRhs[_M21]) + (m[_M33] * pRhs[_M31])) + (m[_M34] * pRhs[_M41]));
    const float m32 = (((m[_M31] * pRhs[_M12]) + (m[_M32] * pRhs[_M22]) + (m[_M33] * pRhs[_M32])) + (m[_M34] * pRhs[_M42]));
    const float m33 = (((m[_M31] * pRhs[_M13]) + (m[_M32] * pRhs[_M23]) + (m[_M33] * pRhs[_M33])) + (m[_M34] * pRhs[_M43]));
    const float m34 = (((m[_M31] * pRhs[_M14]) + (m[_M32] * pRhs[_M24]) + (m[_M33] * pRhs[_M34])) + (m[_M34] * pRhs[_M44]));
    const float m41 = (((m[_M41] * pRhs[_M11]) + (m[_M42] * pRhs[_M21]) + (m[_M43] * pRhs[_M31])) + (m[_M44] * pRhs[_M41]));
    const float m42 = (((m[_M41] * pRhs[_M12]) + (m[_M42] * pRhs[_M22]) + (m[_M43] * pRhs[_M32])) + (m[_M44] * pRhs[_M42]));
    const float m43 = (((m[_M41] * pRhs[_M13]) + (m[_M42] * pRhs[_M23]) + (m[_M43] * pRhs[_M33])) + (m[_M44] * pRhs[_M43]));
    const float m44 = (((m[_M41] * pRhs[_M14]) + (m[_M42] * pRhs[_M24]) + (m[_M43] * pRhs[_M34])) + (m[_M44] * pRhs[_M44]));

    m[_M11] = m11;
    m[_M12] = m12;
    m[_M13] = m13;
    m[_M14] = m14;
    m[_M21] = m21;
    m[_M22] = m22;
    m[_M23] = m23;
    m[_M24] = m24;
    m[_M31] = m31;
    m[_M32] = m32;
    m[_M33] = m33;
    m[_M34] = m34;
    m[_M41] = m41;
    m[_M42] = m42;
    m[_M43] = m43;
    m[_M44] = m44;
    return *this;
  }


  bool Matrix::operator==(const Matrix &rhs) const
  {
    return (m[_M11] == rhs.m[_M11] && m[_M12] == rhs.m[_M12] && m[_M13] == rhs.m[_M13] && m[_M14] == rhs.m[_M14] &&
      m[_M21] == rhs.m[_M21] && m[_M22] == rhs.m[_M22] && m[_M23] == rhs.m[_M23] && m[_M24] == rhs.m[_M24] &&
      m[_M31] == rhs.m[_M31] && m[_M32] == rhs.m[_M32] && m[_M33] == rhs.m[_M33] && m[_M34] == rhs.m[_M34] &&
      m[_M41] == rhs.m[_M41] && m[_M42] == rhs.m[_M42] && m[_M43] == rhs.m[_M43] && m[_M44] == rhs.m[_M44]);
  }


  bool Matrix::operator!=(const Matrix &rhs) const
  {
    return (m[_M11] != rhs.m[_M11] || m[_M12] != rhs.m[_M12] || m[_M13] != rhs.m[_M13] || m[_M14] != rhs.m[_M14] ||
      m[_M21] != rhs.m[_M21] || m[_M22] != rhs.m[_M22] || m[_M23] != rhs.m[_M23] || m[_M24] != rhs.m[_M24] ||
      m[_M31] != rhs.m[_M31] || m[_M32] != rhs.m[_M32] || m[_M33] != rhs.m[_M33] || m[_M34] != rhs.m[_M34] ||
      m[_M41] != rhs.m[_M41] || m[_M42] != rhs.m[_M42] || m[_M43] != rhs.m[_M43] || m[_M44] != rhs.m[_M44]);
  }
}


const Fsl::Matrix operator*(const Fsl::Matrix& lhs, const Fsl::Matrix& rhs)
{
  const float* pLhs = lhs.DirectAccess();
  const float* pRhs = rhs.DirectAccess();
  return Fsl::Matrix(
    (((pLhs[_M11] * pRhs[_M11]) + (pLhs[_M12] * pRhs[_M21])) + (pLhs[_M13] * pRhs[_M31])) + (pLhs[_M14] * pRhs[_M41]),
    (((pLhs[_M11] * pRhs[_M12]) + (pLhs[_M12] * pRhs[_M22])) + (pLhs[_M13] * pRhs[_M32])) + (pLhs[_M14] * pRhs[_M42]),
    (((pLhs[_M11] * pRhs[_M13]) + (pLhs[_M12] * pRhs[_M23])) + (pLhs[_M13] * pRhs[_M33])) + (pLhs[_M14] * pRhs[_M43]),
    (((pLhs[_M11] * pRhs[_M14]) + (pLhs[_M12] * pRhs[_M24])) + (pLhs[_M13] * pRhs[_M34])) + (pLhs[_M14] * pRhs[_M44]),
    (((pLhs[_M21] * pRhs[_M11]) + (pLhs[_M22] * pRhs[_M21])) + (pLhs[_M23] * pRhs[_M31])) + (pLhs[_M24] * pRhs[_M41]),
    (((pLhs[_M21] * pRhs[_M12]) + (pLhs[_M22] * pRhs[_M22])) + (pLhs[_M23] * pRhs[_M32])) + (pLhs[_M24] * pRhs[_M42]),
    (((pLhs[_M21] * pRhs[_M13]) + (pLhs[_M22] * pRhs[_M23])) + (pLhs[_M23] * pRhs[_M33])) + (pLhs[_M24] * pRhs[_M43]),
    (((pLhs[_M21] * pRhs[_M14]) + (pLhs[_M22] * pRhs[_M24])) + (pLhs[_M23] * pRhs[_M34])) + (pLhs[_M24] * pRhs[_M44]),
    (((pLhs[_M31] * pRhs[_M11]) + (pLhs[_M32] * pRhs[_M21])) + (pLhs[_M33] * pRhs[_M31])) + (pLhs[_M34] * pRhs[_M41]),
    (((pLhs[_M31] * pRhs[_M12]) + (pLhs[_M32] * pRhs[_M22])) + (pLhs[_M33] * pRhs[_M32])) + (pLhs[_M34] * pRhs[_M42]),
    (((pLhs[_M31] * pRhs[_M13]) + (pLhs[_M32] * pRhs[_M23])) + (pLhs[_M33] * pRhs[_M33])) + (pLhs[_M34] * pRhs[_M43]),
    (((pLhs[_M31] * pRhs[_M14]) + (pLhs[_M32] * pRhs[_M24])) + (pLhs[_M33] * pRhs[_M34])) + (pLhs[_M34] * pRhs[_M44]),
    (((pLhs[_M41] * pRhs[_M11]) + (pLhs[_M42] * pRhs[_M21])) + (pLhs[_M43] * pRhs[_M31])) + (pLhs[_M44] * pRhs[_M41]),
    (((pLhs[_M41] * pRhs[_M12]) + (pLhs[_M42] * pRhs[_M22])) + (pLhs[_M43] * pRhs[_M32])) + (pLhs[_M44] * pRhs[_M42]),
    (((pLhs[_M41] * pRhs[_M13]) + (pLhs[_M42] * pRhs[_M23])) + (pLhs[_M43] * pRhs[_M33])) + (pLhs[_M44] * pRhs[_M43]),
    (((pLhs[_M41] * pRhs[_M14]) + (pLhs[_M42] * pRhs[_M24])) + (pLhs[_M43] * pRhs[_M34])) + (pLhs[_M44] * pRhs[_M44]));
}
