/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/Exceptions.hpp>
#include <FslBase/Log/Log.hpp>
#include <FslDemoApp/Host/DemoHostFeatureUtil.hpp>
#include <FslDemoHostEGL/EGLDemoHost.hpp>
#include <FslDemoHostEGL/EGLDemoHostOptionParser.hpp>
#include <FslGraphicsEGL/Window/EGLNativeWindowSystemFactory.hpp>
#include <FslGraphicsEGL/Window/IEGLNativeWindowSystem.hpp>
#include <FslGraphicsEGL/Window/IEGLNativeWindow.hpp>
#include <FslDemoApp/Host/DemoAppHostConfigEGL.hpp>
#include <FslGraphics/Window/NativeWindowEventQueue.hpp>
#include <FslGraphics/Window/NativeWindowProcessMessagesArgs.hpp>
#include <FslGraphics/Window/NativeWindowSystemSetup.hpp>
#include <FslGraphicsEGL/EGLCheck.hpp>
#include <FslGraphicsEGL/EGLStringUtil.hpp>
#include <cassert>
#include <iostream>
#include <vector>
#include <EGL/eglext.h>
#include "Service/EGLHost/EGLHostService.hpp"

#if 0
#define LOCAL_LOG(X) FSLLOG("EGLDemoHost: " << X)
#else
#define LOCAL_LOG(X) {}
#endif

#if 1
#define LOCAL_LOG_WARNING(X) FSLLOG_WARNING("EGLDemoHost: " << X)
#else
#define LOCAL_LOG_WARNING(X) {}
#endif

#define EMPTY_VALUE_EGLCONFIG nullptr

namespace Fsl
{
  namespace
  {
    struct FeatureConfig
    {
      bool EnableGLES;
      int ESVersion;

      bool EnableVG;
      int VGVersion;

      DemoHostFeature Feature;

      FeatureConfig()
        : EnableGLES(false)
        , ESVersion(0)
        , EnableVG(false)
        , VGVersion(0)
        , Feature(DemoHostFeatureName::OpenGLES, 0)
      {
      }
    };

    //! @brief copy the supplied EGL config into a deque
    void CopyConfig(std::deque<EGLint>& rEglConfigAttribs, const int32_t* pEglConfigAttribs)
    {
      rEglConfigAttribs.clear();
      if (pEglConfigAttribs != nullptr)
      {
        const int MAX = 1024;
        int count = 0;
        while (*pEglConfigAttribs != EGL_NONE && count < MAX)
        {
          rEglConfigAttribs.push_back(*pEglConfigAttribs);
          ++pEglConfigAttribs;
          ++count;
        }
        if (*pEglConfigAttribs != EGL_NONE)
          throw std::invalid_argument("pEglConfigAttribs seems to be invalid");
        else
          rEglConfigAttribs.push_back(*pEglConfigAttribs);
      }
    }


    void BuildDefaultConfigAttributes(std::deque<EGLint>& rConfigAttributes, const FeatureConfig& featureConfig, const bool bForceConformance)
    {
      rConfigAttributes.push_back(EGL_RED_SIZE);
      rConfigAttributes.push_back(8);

      rConfigAttributes.push_back(EGL_GREEN_SIZE);
      rConfigAttributes.push_back(8);

      rConfigAttributes.push_back(EGL_BLUE_SIZE);
      rConfigAttributes.push_back(8);

      if (featureConfig.EnableGLES)
      {
        switch (featureConfig.ESVersion)
        {
        case 2:
          rConfigAttributes.push_back(EGL_RENDERABLE_TYPE);
          rConfigAttributes.push_back(EGL_OPENGL_ES2_BIT);
          break;
        case 3:
#ifdef EGL_OPENGL_ES3_BIT
          rConfigAttributes.push_back(EGL_RENDERABLE_TYPE);
          rConfigAttributes.push_back(EGL_OPENGL_ES3_BIT);
#elif defined(EGL_OPENGL_ES3_BIT_KHR)
          rConfigAttributes.push_back(EGL_RENDERABLE_TYPE);
          rConfigAttributes.push_back(EGL_OPENGL_ES3_BIT_KHR);
#endif
          break;
        default:
          throw NotImplementedException("Unknown OpenGLES version request");
        }
      }

      if (featureConfig.EnableVG)
      {
        rConfigAttributes.push_back(EGL_RENDERABLE_TYPE);
        rConfigAttributes.push_back(EGL_OPENVG_BIT);

        //  rConfigAttributes.push_back(EGL_ALPHA_SIZE);
        //  rConfigAttributes.push_back(8);
      }

      //if (bUseOpenGLES && bForceConformance)
      //{
      //  rConfigAttributes.push_back(EGL_CONFORMANT);
      //  rConfigAttributes.push_back(EGL_OPENGL_ES2_BIT);
      //}

      // While the default ought to be  EGL_SURFACE_TYPE = EGL_WINDOW_BIT
      // The arm mali 1.3.2 and 1.4 emulator fails to render correctly without this on newer nvidia cards.
      rConfigAttributes.push_back(EGL_SURFACE_TYPE);
      rConfigAttributes.push_back(EGL_WINDOW_BIT);

      rConfigAttributes.push_back(EGL_SAMPLES);
      rConfigAttributes.push_back(0);

      rConfigAttributes.push_back(EGL_NONE);
    }


    const Point2 GetEGLScreenResolution(EGLDisplay hEglDisplay, EGLSurface hEglSurface)
    {
      EGLint windowWidth;
      EGLint windowHeight;
      eglQuerySurface(hEglDisplay, hEglSurface, EGL_WIDTH, &windowWidth);
      eglQuerySurface(hEglDisplay, hEglSurface, EGL_HEIGHT, &windowHeight);
      return Point2(windowWidth, windowHeight);
    }


    //! @brief replace the attribute if its found
    //! @return true if replaced, false if not found
    bool ReplaceAttribute(std::deque<EGLint>& rConfigAttributes, const EGLint key, const EGLint value)
    {
      assert((rConfigAttributes.size() % 2) == 1 && rConfigAttributes.back() == EGL_NONE);

      std::deque<EGLint>::iterator itr = rConfigAttributes.begin();
      int index = 0;
      while (itr != rConfigAttributes.end() && *itr != EGL_NONE)
      {
        if ((index & 1) == 0 && *itr == key)
        {
          ++itr;
          *itr = value;
          return true;
        }
        ++itr;
        ++index;
      }
      return false;
    }

    void MergeWithOverwriteConfigAttributes(std::deque<EGLint>& rConfigAttributes, const std::deque<EGLint>& srcConfigAttribs)
    {
      if (srcConfigAttribs.size() == 0)
        return;

      assert((rConfigAttributes.size() % 2) == 1);
      assert((srcConfigAttribs.size() % 2) == 1);

      std::deque<EGLint>::const_iterator itr = srcConfigAttribs.begin();
      while (itr != srcConfigAttribs.end() && *itr != EGL_NONE)
      {
        EGLint key = *itr;
        ++itr;
        EGLint value = *itr;
        ++itr;

        if (!ReplaceAttribute(rConfigAttributes, key, value))
        {
          rConfigAttributes.pop_back();
          rConfigAttributes.push_back(key);
          rConfigAttributes.push_back(value);
          rConfigAttributes.push_back(EGL_NONE);
        }
      }

    }

    void ApplyHostOverrides(std::deque<EGLint>& rConfigAttributes, const std::shared_ptr<EGLDemoHostOptionParser> options)
    {
      std::deque<EGLint> forceConfigAttributes;
      options->ExtractConfigAttributes(forceConfigAttributes);
      if (forceConfigAttributes.size() > 0)
        MergeWithOverwriteConfigAttributes(rConfigAttributes, forceConfigAttributes);
    }

    const FeatureConfig ExamineFeatureRequest(const std::shared_ptr<DemoHostAppSetup::DemoHostFeatureDeque>& features)
    {
      FeatureConfig featureConfig;
      DemoHostAppSetup::DemoHostFeatureDeque::const_iterator itr = features->begin();
      while (itr != features->end())
      {
        if (itr->Name == DemoHostFeatureName::OpenGLES)
        {
          featureConfig.EnableGLES = true;

          int major, minor;
          DemoHostFeatureUtil::DecodeOpenGLESVersion(itr->Version, major, minor);
          featureConfig.ESVersion = major;
          featureConfig.Feature = *itr;
        }
        else if (itr->Name == DemoHostFeatureName::OpenVG)
        {
          featureConfig.EnableVG = true;

          int major, minor;
          DemoHostFeatureUtil::DecodeOpenVGVersion(itr->Version, major, minor);
          featureConfig.VGVersion = major;
          featureConfig.Feature = *itr;
        }
        ++itr;
      }
      return featureConfig;
    }

    void ExamineFinalConfig(const std::vector<EGLint>& finalConfigAttribs)
    {
      // start at 1 so that the size check is simpler
      for (size_t i = 1; i < finalConfigAttribs.size(); i += 2)
      {
        switch (finalConfigAttribs[i - 1])
        {
        case EGL_RED_SIZE:
        case EGL_GREEN_SIZE:
        case EGL_BLUE_SIZE:
        case EGL_ALPHA_SIZE:
          if (finalConfigAttribs[i] < 0)
          {
            FSLLOG_WARNING(EGLStringUtil::GetConfigEnumToString(finalConfigAttribs[i - 1]) << " can not be negative.");
          }
          break;
        }
      }
    }

    void DoLogConfig(const std::vector<EGLint>& finalConfigAttribs)
    {
      // start at 1 so that the size check is simpler
      FSLLOG("Requested EGL config");
      for (size_t i = 1; i < finalConfigAttribs.size(); i += 2)
      {
        FSLLOG(EGLStringUtil::GetConfigEnumToString(finalConfigAttribs[i - 1]) << "=" << finalConfigAttribs[i]);
      }
    }

    void DoLogConfigComparison(EGLDisplay hDisplay, EGLConfig config, const std::vector<EGLint>& finalConfigAttribs)
    {
      FSLLOG("Final vs requested EGL config");

      for (size_t i = 1; i < finalConfigAttribs.size(); i += 2)
      {
        const EGLint attrib = finalConfigAttribs[i - 1];
        const EGLint requestedValue = finalConfigAttribs[i];
        EGLint actualValue;
        EGL_CHECK(eglGetConfigAttrib(hDisplay, config, attrib, &actualValue));

        if (actualValue != requestedValue)
        {
          FSLLOG(EGLStringUtil::GetConfigEnumToString(finalConfigAttribs[i - 1]) << " is " << actualValue << " requested " << requestedValue);
        }
      }
    }
  }


  EGLDemoHost::EGLDemoHost(const DemoHostConfig& demoHostConfig)
    : ADemoHost(demoHostConfig)
    , m_demoHostConfig(demoHostConfig)
    , m_finalConfigAttribs()
    , m_eglContextClientVersion(0)
    , m_nativeWindowSetup()
    , m_windowSystem()
    , m_window()
    , m_hostService(demoHostConfig.GetServiceProvider().Get<EGLHostService>())
    , m_hDisplay(EGL_NO_DISPLAY)
    , m_hSurface(EGL_NO_SURFACE)
    , m_hContext(EGL_NO_CONTEXT)
    , m_hConfig(EMPTY_VALUE_EGLCONFIG)
    , m_isActivated(true)
    , m_enableGLES(false)
    , m_enableVG(false)
    , m_logSelectedConfig(false)
    , m_activeApi(DemoHostFeatureName::OpenGLES, 0)
  {
    // - GetOptions checks for nullptr's so we dont have to
    const std::shared_ptr<EGLDemoHostOptionParser> options = demoHostConfig.GetOptions<EGLDemoHostOptionParser>();

    const NativeWindowSystemSetup nativeWindowSystemSetup(demoHostConfig.GetEventQueue(), options->GetNativeWindowConfig(), options->GetNativeWindowTag());
    m_windowSystem = EGLNativeWindowSystemFactory::Allocate(nativeWindowSystemSetup);

    const DemoHostAppSetup hostAppSetup = demoHostConfig.GetDemoHostAppSetup();
    // Retrieve the custom config
    const std::shared_ptr<DemoAppHostConfigEGL> appHostConfig = hostAppSetup.GetDemoAppHostConfig<DemoAppHostConfigEGL>();

    // Check that its a OpenGLES demo app
    FeatureConfig featureConfig = ExamineFeatureRequest(hostAppSetup.DemoHostFeatures);
    if (!featureConfig.EnableGLES && !featureConfig.EnableVG)
      throw UsageErrorException("The EGLDemo host currently expects that OpenGLES or OpenVG has to be enabled");
    m_activeApi = featureConfig.Feature;
    m_enableGLES = featureConfig.EnableGLES;
    m_enableVG = featureConfig.EnableVG;

    m_eglContextClientVersion = featureConfig.ESVersion;

    // Do a basic validation on the app's custom EGL config attribs
    std::deque<EGLint> eglConfigAttribs;
    CopyConfig(eglConfigAttribs, appHostConfig->GetEglConfigAttribs());

    const ConfigControl::Enum configControl = appHostConfig->GetConfigControl();
    if (!(eglConfigAttribs.size() == 0 || ((eglConfigAttribs.size() % 2) == 1 && eglConfigAttribs.back() == EGL_NONE)))
      throw std::invalid_argument("eglConfigAttribs is in a unexpected format");

    // Then build our default egl config
    std::deque<EGLint> configAttribs;
    if (eglConfigAttribs.size() <= 0 || configControl == ConfigControl::Overwrite)
      BuildDefaultConfigAttributes(configAttribs, featureConfig, false);
    if (eglConfigAttribs.size() > 0)
      MergeWithOverwriteConfigAttributes(configAttribs, eglConfigAttribs);

    // Allow the command line options to override things in the config
    ApplyHostOverrides(configAttribs, options);
    
    // Store the final configuration for later use
    m_finalConfigAttribs.resize(configAttribs.size());
    std::copy(configAttribs.begin(), configAttribs.end(), m_finalConfigAttribs.begin());

    ExamineFinalConfig(m_finalConfigAttribs);
    if (options->IsLogConfigEnabled())
    {
      DoLogConfig(m_finalConfigAttribs);
      m_logSelectedConfig = true;
    }

    // Prepare the native window setup
    m_nativeWindowSetup.reset(new NativeWindowSetup(demoHostConfig.GetEventQueue(), options->GetNativeWindowConfig()));

    // Prepare everything
    Init();
  }


  EGLDemoHost::~EGLDemoHost()
  {
    LOCAL_LOG("Destroying");
    Shutdown();
    LOCAL_LOG("Destroyed");
  }


  void EGLDemoHost::OnActivate()
  {
    m_isActivated = true;
  }


  void EGLDemoHost::OnDeactivate()
  {
    m_isActivated = false;
  }


  void EGLDemoHost::OnSuspend()
  {
    Shutdown();
  }


  void EGLDemoHost::OnResume()
  {
    Init();
  }


  DemoHostFeature EGLDemoHost::GetActiveAPI() const
  {
    return m_activeApi;
  }


  Point2 EGLDemoHost::GetScreenResolution() const
  {
    assert(! m_enableGLES || m_eglContextClientVersion > 0);
    // The EGL state should be 100% ready
    assert(m_hDisplay != EGL_NO_DISPLAY);
    assert(m_hContext != EGL_NO_CONTEXT);
    assert(m_hSurface != EGL_NO_SURFACE);
    assert(m_hConfig != EMPTY_VALUE_EGLCONFIG);
    assert(m_window);

    return GetEGLScreenResolution(m_hDisplay, m_hSurface);
  }


  bool EGLDemoHost::SwapBuffers()
  {
    if (!m_isActivated)
      return true;

    // Validate that we are not suspended
    assert(m_window);
    assert(m_hDisplay != EGL_NO_DISPLAY);
    assert(m_hContext != EGL_NO_CONTEXT);
    assert(m_hSurface != EGL_NO_SURFACE);
    assert(m_hConfig != EMPTY_VALUE_EGLCONFIG);
    return eglSwapBuffers(m_hDisplay, m_hSurface) == EGL_TRUE;
  }


  bool EGLDemoHost::ProcessNativeMessages(const bool allowBlock)
  {
    // This will be called while suspended
    // Allow the native window to process messages
    NativeWindowProcessMessagesArgs args(allowBlock);
    return m_windowSystem->ProcessMessages(args);
  }


  void EGLDemoHost::Init()
  {
    LOCAL_LOG("Init");
    try
    {
      // Get the basic GL configuration ready
      InitEGL();

      // Create the main surface and context
      InitSurfaceAndContext();
    }
    catch (const std::exception&)
    {
      //LOCAL_LOG("Init failed with exception: " << ex.what());
      Shutdown();
      throw;
    }
  }


  void EGLDemoHost::Shutdown()
  {
    LOCAL_LOG("Shutdown");
    ShutdownSurfaceAndContext();
    ShutdownEGL();
  }


  void EGLDemoHost::InitEGL()
  {
    assert(m_nativeWindowSetup);
    assert(!m_window);
    assert(m_hDisplay == EGL_NO_DISPLAY);
    assert(m_hContext == EGL_NO_CONTEXT);
    assert(m_hSurface == EGL_NO_SURFACE);
    assert(m_hConfig == EMPTY_VALUE_EGLCONFIG);

    // Acquire the various native display handles
    EGLNativeDisplayType hDisplay = m_windowSystem->GetDisplayType();
    try
    {
      m_hDisplay = EGL_CHECK(eglGetDisplay(hDisplay));

      LOCAL_LOG("Initialize");

      // Configure the display
      EGL_CHECK(eglInitialize(m_hDisplay, nullptr, nullptr));
      
      if (m_enableVG)
      {
        LOCAL_LOG("Binding VG");
        EGL_CHECK(eglBindAPI(EGL_OPENVG_API));
      }
      if (m_enableGLES)
      {
        LOCAL_LOG("Binding GLES");
        EGL_CHECK(eglBindAPI(EGL_OPENGL_ES_API));
      }

      LOCAL_LOG("Choosing config");

      EGLint numConfigs;
      EGL_CHECK(eglChooseConfig(m_hDisplay, &m_finalConfigAttribs[0], &m_hConfig, 1, &numConfigs));
      
      if (numConfigs != 1)
        throw EGLGraphicsException("The required configuration was not supported.");

      if (m_logSelectedConfig)
        DoLogConfigComparison(m_hDisplay, m_hConfig, m_finalConfigAttribs);

      LOCAL_LOG("Creating native window");

      // Prepare the native window
      const NativeEGLSetup nativeEglSetup(hDisplay, m_hDisplay, m_hConfig);
      m_window = m_windowSystem->CreateNativeWindow(*m_nativeWindowSetup, nativeEglSetup);
    }
    catch (const std::exception&)
    {
      ShutdownEGL();
      throw;
    }
  }


  void EGLDemoHost::InitSurfaceAndContext()
  {
    //! The app manager should not exist when this is called

    assert(!m_enableGLES || m_eglContextClientVersion > 0);
    assert(m_hDisplay != EGL_NO_DISPLAY);
    assert(m_hConfig != EMPTY_VALUE_EGLCONFIG);
    assert(m_window);
    if (m_hDisplay == EGL_NO_DISPLAY || m_hConfig == EMPTY_VALUE_EGLCONFIG || !m_window)
      return;

    try
    {
      EGLint contextAttribListES[] = { EGL_CONTEXT_CLIENT_VERSION, m_eglContextClientVersion, EGL_NONE };
      EGLint contextAttribListVG[] = { EGL_NONE };
      EGLint* contextAttribList = (m_enableGLES ? contextAttribListES : contextAttribListVG);      

      const EGLNativeWindowType hWindow = m_window->GetWindowType();

      // Setup the rest
      LOCAL_LOG("Creating window surface");
      m_hSurface = EGL_CHECK(eglCreateWindowSurface(m_hDisplay, m_hConfig, hWindow, nullptr));

      LOCAL_LOG("Creating context");
      m_hContext = EGL_CHECK(eglCreateContext(m_hDisplay, m_hConfig, EGL_NO_CONTEXT, contextAttribList));

      LOCAL_LOG("Making current");
      EGL_CHECK(eglMakeCurrent(m_hDisplay, m_hSurface, m_hSurface, m_hContext));
    }
    catch (const std::exception&)
    {
      ShutdownSurfaceAndContext();
      throw;
    }
  }


  void EGLDemoHost::ShutdownSurfaceAndContext()
  {
    if (m_hDisplay == EGL_NO_DISPLAY)
      return;

    assert(m_hConfig != EMPTY_VALUE_EGLCONFIG);
    assert(m_window);

    EGL_LOG_ERROR(eglMakeCurrent(m_hDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT));

    assert(eglGetError() == EGL_SUCCESS);
    if (m_hContext != EGL_NO_CONTEXT)
    {
      LOCAL_LOG("Destroying context");
      m_hostService->SetContext(EGL_NO_CONTEXT);
      // NOTE: ARM-EMU - For some reason this gives a error on shutdown
      EGL_LOG_ERROR(eglDestroyContext(m_hDisplay, m_hContext));
      m_hContext = EGL_NO_CONTEXT;
    }
    
    if (m_hSurface != EGL_NO_SURFACE)
    {
      LOCAL_LOG("Destroying surface");
      m_hostService->SetSurface(EGL_NO_SURFACE);
      EGL_LOG_ERROR(eglDestroySurface(m_hDisplay, m_hSurface));
      m_hSurface = EGL_NO_SURFACE;
    }
  }


  void EGLDemoHost::ShutdownEGL()
  {
    //! The app manager should have been released before calling this
    assert(m_hContext == EGL_NO_CONTEXT);
    assert(m_hSurface == EGL_NO_SURFACE);

    if (m_hDisplay != EGL_NO_DISPLAY)
    {
      LOCAL_LOG("Destroying window");
      m_hostService->SetDisplay(EGL_NO_DISPLAY);
      // Release the native window
      m_window.reset();

      m_hConfig = EMPTY_VALUE_EGLCONFIG;
      LOCAL_LOG("Terminating egl");
      eglTerminate(m_hDisplay);
      assert(eglGetError() == EGL_SUCCESS);
      m_hDisplay = EGL_NO_DISPLAY;
      eglReleaseThread();
    }
  }
}
