#ifdef FSL_ENABLE_DEVIL
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/Log/Log.hpp>
#include <FslBase/Exceptions.hpp>
#include <FslBase/Noncopyable.hpp>
#include <FslBase/IO/File.hpp>
#include <FslBase/IO/Path.hpp>
#include <FslGraphics/BitmapUtil.hpp>
#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/PixelFormatUtil.hpp>
#include <FslGraphics/IO/BMPUtil.hpp>
#include <FslDemoPlatform/Service/Image/ImageServiceDevIL.hpp>
#include <cassert>
#include <sstream>
#include <IL/il.h>

namespace Fsl
{
  namespace
  {
    // TODO: validate the BPP 
    PixelFormat::Enum Convert(ILint format)
    {
      switch (format)
      {
      case IL_RGB:
        return PixelFormat::RGB888;
      case IL_RGBA:
        return PixelFormat::RGBA8888;
      case IL_BGR:
        return PixelFormat::BGR888;
      case IL_BGRA:
        return PixelFormat::BGRA8888;
      case IL_LUMINANCE:
        return PixelFormat::LUMINANCE8;
      case IL_LUMINANCE_ALPHA:
        return PixelFormat::LUMINANCE_ALPHA88;
      default:
        throw GraphicsException("Unsupported image format");
      }
    }

    ILint Convert(PixelFormat::Enum format)
    {
      switch (format)
      {
      case PixelFormat::RGB888:
        return IL_RGB;
      case PixelFormat::RGBA8888:
        return IL_RGBA;
      case PixelFormat::BGR888:
        return IL_BGR;
      case PixelFormat::BGRA8888:
        return IL_BGRA;
      case PixelFormat::LUMINANCE8:
        return IL_LUMINANCE;
      case PixelFormat::LUMINANCE_ALPHA88:
        return IL_LUMINANCE_ALPHA;
      default:
        throw UnsupportedPixelFormatException(format);
      }
    }

    struct ScopedDevILImage : private Noncopyable
    {
      ILuint Id;

      ScopedDevILImage()
      {
        ilGenImages(1, &Id);
        if (ilGetError() != IL_NO_ERROR)
          throw GraphicsException("Failed to create DevIL image");
      }
      ~ScopedDevILImage()
      {
        ilDeleteImages(1, &Id);
      }
    };


    bool LoadILImage(Bitmap& rBitmap, const IO::Path& path, ILint desiredFormat, const bool convertFormat)
    {
      ILuint devilError = IL_NO_ERROR;
      ScopedDevILImage image;

      ilBindImage(image.Id);
      // FIX: investigate if DevIL supports UTF8 on all platforms.
      // Since the ToAsciiString() conversion breaks UTF8 support.
      // Its quite likely that UTF8 works on linux based platforms, 
      // but windows might require a different solution.

#ifdef _WIN32
      // Loads into the current bound image
      ilLoadImage(path.ToWString().c_str());
#else
      ilLoadImage(path.ToAsciiString().c_str());
#endif

      devilError = ilGetError();
      if (devilError != IL_NO_ERROR)
      {
        FSLLOG_WARNING("devIL image loading not successfull.\n");
        return false;
      }

      int width = ilGetInteger(IL_IMAGE_WIDTH);
      int height = ilGetInteger(IL_IMAGE_HEIGHT);
      //int bpp = ilGetInteger(IL_IMAGE_BITS_PER_PIXEL);
      int format = ilGetInteger(IL_IMAGE_FORMAT);

      if (convertFormat && format != desiredFormat)
      {
        ilConvertImage(desiredFormat, IL_UNSIGNED_BYTE);
      }
      format = ilGetInteger(IL_IMAGE_FORMAT);


      // Extract the image. 
      {
        // TODO: Remove the temporary content array by copying it into the bitmap directly instead
        const PixelFormat::Enum currentPixelFormat = Convert(format);
        const int bytesPerPixel = PixelFormatUtil::GetBytesPerPixel(currentPixelFormat);
        std::vector<uint8_t> content(width * height * bytesPerPixel);
        ilCopyPixels(0, 0, 0, width, height, 1, format, IL_UNSIGNED_BYTE, content.data());
        rBitmap.Reset(content.data(), width, height, currentPixelFormat);
      }

      devilError = ilGetError();

      // Log any error that occurs
      FSLLOG_WARNING_IF(devilError != IL_NO_ERROR, "devIL image conversion not successfull (" << devilError << ").\n");
      return (devilError == IL_NO_ERROR);
    }
  }


  ImageServiceDevIL::ImageServiceDevIL(const ServiceProvider& serviceProvider)
    : ThreadLocalService(serviceProvider)
  {
    ILuint devilError = IL_NO_ERROR;
    ilInit();
    devilError = ilGetError();
    if (devilError != IL_NO_ERROR)
    {
      throw std::runtime_error("devIL lib init failed");
    }

    // Make sure the images are using the expected origin
    ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
    ilEnable(IL_ORIGIN_SET);
  }


  ImageServiceDevIL::~ImageServiceDevIL()
  {
  }


  bool ImageServiceDevIL::TryRead(Bitmap& rBitmap, const IO::Path& path)
  {
    try
    {
      return LoadILImage(rBitmap, path, 0, false);
    }
    catch (std::exception&ex)
    {
      FSLLOG_WARNING("devIL image conversion not successfull (" << ex.what() << ").\n");
      return false;
    }
  }


  bool ImageServiceDevIL::TryRead(Bitmap& rBitmap, const IO::Path& path, const PixelFormat::Enum desiredPixelFormat)
  {
    try
    {
      return LoadILImage(rBitmap, path, Convert(desiredPixelFormat), true);
    }
    catch (std::exception&ex)
    {
      FSLLOG_WARNING("devIL image conversion not successfull (" << ex.what() << ").\n");
      return false;
    }
  }
}
#endif
