/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/Exceptions.hpp>
#include <FslBase/Log/Log.hpp>
#include <FslDemoPlatform/Service/MMDCStats/MMDCStatsService.hpp>
#include <FslBase/System/Platform/SystemMemoryAccess.hpp>
#include <cassert>

namespace Fsl
{
  namespace
  {
    bool ProfileMMDCAXIBusBegin(const HighResolutionTimer& timer, uint64_t& rSampleStartTime)
    {
      bool ret = true;
      try
      {
        //MADPCR1 (0x021B0414), set AXI_ID to 0 and AXI ID MASK to 0
        //meaning all access to memory is profiled
        SystemMemoryAccess::SetUInt32(0x021B0414, 0x0);

        rSampleStartTime = timer.GetTime();

        //MADPCR0 (0x021B0410), set profiling enable bit[0] (others ok at zero)
        SystemMemoryAccess::SetUInt32(0x021B0410, 0x00000001);
      }
      catch (const std::exception& ex)
      {
        FSLLOG_ERROR("caught exception during register readout" << ex.what()); //minor fault, continue execution
        ret = false;
      }
      return ret;
    }


    //interval between begin and end should be long enough to grab a stat without the counter overflowing
    bool ProfileMMDCAXIBusEnd(const HighResolutionTimer& timer, const uint64_t sampleStartTime, uint32_t& rReadPerSec, uint32_t& rWritePerSec, uint32_t& rTotalBytes, float& rLoad)
    {
      uint32_t readCount;
      uint32_t writeCount;
      uint32_t cycleCount;
      uint32_t busyCount;
      bool ret = true;
      try
      {
        //grab the MADPCR before freezing the profiler to check the overflow bit
        uint32_t madpcr0 = SystemMemoryAccess::GetUInt32(0x021B0410);

        //MADPCR0 (0x021B0410), set profiling freeze bit[2] (enable on bit[0] others ok at zero)
        SystemMemoryAccess::SetUInt32(0x021B0410, 0x00000005);

        const uint64_t sampleTimeStop = timer.GetTime();

        if (madpcr0 & 0x8)
        {
          //counter overflow detected via MADPCR0 bit[4], clear the profiler and ignore invalid data
          SystemMemoryAccess::SetUInt32(0x021B0410, 0x2);
          SystemMemoryAccess::SetUInt32(0x021B0410, 0x0);
          SystemMemoryAccess::SetUInt32(0x021B0414, 0x0);
          ret = false;
        }
        else
        {
          //MADPSR0, cycle count
          cycleCount = SystemMemoryAccess::GetUInt32(0x021B0418);
          //MADPSR1, busy count
          busyCount = SystemMemoryAccess::GetUInt32(0x021B041C);
          //MADPSR4, bytes read
          readCount = SystemMemoryAccess::GetUInt32(0x021B0428);
          //MADPSR5, bytes written
          writeCount = SystemMemoryAccess::GetUInt32(0x021B042C);

          //writing to MADPCR0 bit[1] clears the profiling counters
          SystemMemoryAccess::SetUInt32(0x021B0410, 0x2);
          //clear the AXI ID and profiler controls
          SystemMemoryAccess::SetUInt32(0x021B0410, 0x0);
          SystemMemoryAccess::SetUInt32(0x021B0414, 0x0);

          uint64_t deltaTime = (sampleTimeStop - sampleStartTime);
          if (deltaTime > 0)
          {
            rReadPerSec = static_cast<int32_t>((readCount / deltaTime) * 1000);
            rWritePerSec = static_cast<int32_t>((writeCount / deltaTime) * 1000);
          }
          else
          {
            rReadPerSec = 0;
            rWritePerSec = 0;
          }

          rTotalBytes = readCount + writeCount;
          if (cycleCount != 0)
            rLoad = ((float)busyCount) / cycleCount;
          else
            rLoad = 0.0f;
          //        std::cout << "Raw Rb:" << readCount << " Wb: " << writeCount << " Dt: " << (timeStop-timeStart) << "\n";
        }
      }
      catch (const std::exception& ex)
      {
        FSLLOG_ERROR("caught exception during register readout" << ex.what()); //minor fault, continue execution
        ret = false;
      }
      return ret;
    }
  }

  MMDCStatsService::MMDCStatsService(const ServiceProvider& serviceProvider)
    : ThreadLocalService(serviceProvider)
    , m_timer()
    , m_profiler(serviceProvider.Get<IProfilerService>())
    //, m_handleGPU3D(m_profiler, m_profiler->CreateCustomCounter("%_3d", 0, 1000, Color(0xFF70DB93)))
    //, m_handleGPU2D(m_profiler, m_profiler->CreateCustomCounter("%_2d", 0, 1000, Color(0xFF70DB93)))
    //, m_handleGPUVG(m_profiler, m_profiler->CreateCustomCounter("%_vg", 0, 1000, Color(0xFF70DB93)))
    , m_handleMMDC()
    , m_handleReads()
    , m_handleWrites()
    , m_enabled(false)
    , m_initialized(false)
    , m_crashed(false)
    , m_sampleStartTime(0)
    , m_mmdcReadsPerSec(0)
    , m_mmdcWritesPerSec(0)
    , m_mmdcTotalBytes(0)
    , m_mmdcLoad(0)
  {

  }


  MMDCStatsService::~MMDCStatsService()
  {
  }


  void MMDCStatsService::Update()
  {
    ThreadLocalService::Update();

    if (! m_enabled || m_crashed)
      return;

    if (m_initialized)
    {
      if (ProfileMMDCAXIBusEnd(m_timer, m_sampleStartTime, m_mmdcReadsPerSec, m_mmdcWritesPerSec, m_mmdcTotalBytes, m_mmdcLoad))
      {
        // got new stats
        //FSLLOG("Reads: " << m_mmdcReadsPerSec << " Writes: " << m_mmdcWritesPerSec << " Total: " << m_mmdcTotalBytes << " load: " << m_mmdcLoad);
        int32_t loadPercentage = static_cast<int32_t>(m_mmdcLoad * 100.0f);
        m_profiler->Set(m_handleMMDC, loadPercentage);
        m_profiler->Set(m_handleReads, m_mmdcReadsPerSec);
        m_profiler->Set(m_handleWrites, m_mmdcWritesPerSec);
      }
    }
    
    if (ProfileMMDCAXIBusBegin(m_timer, m_sampleStartTime))
    {
      m_initialized = true;
    }
    //else
    //{
    //  m_crashed = true;
    //  m_initialized = false;
    //}
  }


  void MMDCStatsService::SetEnabled(const bool enabled)
  {
    if (enabled != m_enabled)
    {
      m_enabled = enabled;
      if (!enabled)
      {
        m_handleMMDC.Reset();
        m_handleReads.Reset();
        m_handleWrites.Reset();
      }
      else
      {
        m_handleMMDC.Reset(m_profiler, m_profiler->CreateCustomCounter("%_mmdc", 0, 100, Color(0xFF70DB93)));
        m_handleReads.Reset(m_profiler, m_profiler->CreateCustomCounter("r", 0, 1000000, Color(0xFFFFAE00)));
        m_handleWrites.Reset(m_profiler, m_profiler->CreateCustomCounter("w", 0, 1000000, Color(0xFF00FF48)));
      }
    }
  }

}
