#ifndef FSLGRAPHICS_BITMAP_HPP
#define FSLGRAPHICS_BITMAP_HPP
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/Noncopyable.hpp>
#include <FslGraphics/RawBitmap.hpp>
#include <FslGraphics/RawBitmapEx.hpp>
#include <FslGraphics/BitmapClearMethod.hpp>
#include <FslGraphics/StrideRequirement.hpp>
#include <vector>

namespace Fsl
{

  //@note All the methods that will default to using StrideRequirement::Any will use the minimum stride for allocation of the bitmap.
  //      The none part is just set to mean that there are no special requirements that should be maintained for the bitmap if its
  //      modified.
  class Bitmap
  {
    //! The raw image data
    std::vector<uint32_t> m_content;
    int32_t m_width;
    int32_t m_height;
    int32_t m_stride;
    PixelFormat::Enum m_pixelFormat;
    StrideRequirement::Enum m_strideRequirement;
    mutable bool m_bIsLocked;
  public:
    //! @brief Create a empty bitmap
    Bitmap();
    //! @brief Create a empty bitmap of the given width, height and pixel format (The default stride preference is StrideRequirement::Any)
    Bitmap(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat);
    //! @brief Create a empty bitmap of the given width, height, pixel format and stride (The default stride preference is StrideRequirement::Any)
    Bitmap(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The input bitmap is assumed to use the minimum stride and the created bitmap will use StrideRequirement::Any to find its stride.
    Bitmap(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The created bitmap will use StrideRequirement::Any to find its stride.
    Bitmap(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The created bitmap will use StrideRequirement::Any to find its stride.
    Bitmap(const RawBitmap& srcBitmap);

    //! The width of the bitmap in pixels
    int32_t Width() const;

    //! The height of the bitmap in pixels
    int32_t Height() const;

    //! The stride of the bitmap in bytes
    int32_t Stride() const;

    //! Get the pixel format of the raw bitmap
    PixelFormat::Enum GetPixelFormat() const;

    //! @brief Get the bitmap stride preference.
    StrideRequirement::Enum GetStrideRequirement() const;

    //! @brief Get the stride that this bitmap would prefer given it had to be stored in the given pixel format.
    //! @note  This takes into account the current StrideRequirement of the bitmap.
    int32_t GetPreferredStride(const PixelFormat::Enum pixelFormat) const;

    //! @brief Reset the image to the given dimensions.
    //! @param clearMethod defines what to do with the existing content
    //! @note This resets the stride preference to StrideRequirement::Any
    void Reset(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const BitmapClearMethod::Enum clearMethod);

    //! @brief Reset the image to the given dimensions.
    //! @param clearMethod defines what to do with the existing content
    //! @note This resets the stride preference to StrideRequirement::Any
    void Reset(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride, const BitmapClearMethod::Enum clearMethod);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The input bitmap is assumed to use the minimum stride and the created bitmap will use StrideRequirement::Any to find its stride.
    //! @note This resets the stride preference to StrideRequirement::Any
    void Reset(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The created bitmap will use StrideRequirement::Any to find its stride.
    //! @note This resets the stride preference to StrideRequirement::Any
    void Reset(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride);

    //! @brief Create a bitmap from the supplied content (the content will be copied). 
    //         The created bitmap will use StrideRequirement::Any to find its stride.
    //! @note This resets the stride preference to StrideRequirement::Any
    void Reset(const RawBitmap& srcBitmap);

    //! @brief set the pixel at the given coordinate (if the coordinate is outside the bitmap nothing happens)
    //! @note This is a very slow operation
    void SetNativePixel(const int x, const int y, const uint32_t color);

    //! @brief read the pixel at the given coordinate (if the coordinate is outside the bitmap zero is returned)
    //! @note This is a very slow operation
    uint32_t GetNativePixel(const int x, const int y) const;

    //! Provides direct access to the scoped bitmap during its lifetime.
    class ScopedDirectAccess : private Noncopyable
    {
      const Bitmap* m_pBitmap1;
      Bitmap* m_pBitmap2;
    public:
      // Read only lock
      ScopedDirectAccess(const Bitmap& bitmap, RawBitmap& rRawBitmap)
        : m_pBitmap1(&bitmap)
        , m_pBitmap2(nullptr)
      {
        rRawBitmap = bitmap.Lock();
      }

      // Read/write lock
      ScopedDirectAccess(Bitmap& bitmap, RawBitmapEx& rRawBitmap)
        : m_pBitmap1(nullptr)
        , m_pBitmap2(&bitmap)
      {
        rRawBitmap = bitmap.LockEx();
      }

      ~ScopedDirectAccess()
      {
        if (m_pBitmap1 != nullptr)
          m_pBitmap1->Unlock();
        else if (m_pBitmap2 != nullptr)
          m_pBitmap2->Unlock();
      }
    };
  private:
    RawBitmap Lock() const;
    RawBitmapEx LockEx();
    void Unlock();
    void Unlock() const;
    void ResizeToFit(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const StrideRequirement::Enum strideRequirement, const int32_t stride = 0);
    void Clear(const BitmapClearMethod::Enum clearMethod);
  };
}

#endif


