/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/Bitmap.hpp>
#include <FslGraphics/PixelFormatUtil.hpp>
#include <FslGraphics/RawBitmapUtil.hpp>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstring>

namespace Fsl
{
  namespace
  {
    void Copy(RawBitmapEx& dstBitmap, const RawBitmap& srcBitmap)
    {
      assert(srcBitmap.IsValid());
      assert(dstBitmap.IsValid());
      assert(dstBitmap.Content() != srcBitmap.Content());
      assert(dstBitmap.Width() == srcBitmap.Width());
      assert(dstBitmap.Height() == srcBitmap.Height());
      assert(dstBitmap.GetPixelFormat() == srcBitmap.GetPixelFormat());

      if (dstBitmap.Stride() == srcBitmap.Stride())
      {
        // The stride is identical -> so just do a direct copy
        const int32_t cbBitmap = srcBitmap.Height() * srcBitmap.Stride();
        memcpy(dstBitmap.Content(), srcBitmap.Content(), cbBitmap);
      }
      else
      {
        // Slow generic fall-back copy to handle mixed stride images
        const uint32_t minimumStride = PixelFormatUtil::CalcMinimumStride(dstBitmap.Width(), dstBitmap.GetPixelFormat());

        const int32_t dstStride = dstBitmap.Stride();
        const uint8_t* pSrc = static_cast<const uint8_t*>(srcBitmap.Content());
        uint8_t* pDst = static_cast<uint8_t*>(dstBitmap.Content());
        const uint8_t*const pDstEnd = pDst + (dstStride * dstBitmap.Height());
        const int32_t srcAdd = srcBitmap.Stride() - minimumStride;
        const int32_t dstAdd = dstStride - minimumStride;

        uint8_t* pDstTmp = pDst + minimumStride;
        while (pDst < pDstEnd)
        {
          while (pDst < pDstTmp)
          {
            *pDst = *pSrc;
            ++pSrc;
            ++pDst;
          }
          pSrc += srcAdd;
          pDst += dstAdd;
          pDstTmp += dstStride;
        }
      }
    }
  }

  Bitmap::Bitmap()
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(PixelFormat::RGBA8888)
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
  }


  Bitmap::Bitmap(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat)
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(pixelFormat)
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
    ResizeToFit(width, height, pixelFormat, m_strideRequirement);
  }


  Bitmap::Bitmap(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride)
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(pixelFormat)
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
    if (stride < 0)
      throw std::invalid_argument("stride can not be < 0");

    ResizeToFit(width, height, pixelFormat, m_strideRequirement, stride);
  }


  Bitmap::Bitmap(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat)
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(pixelFormat)
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
    if (pContent == nullptr)
      throw std::invalid_argument("pContent can not be null");

    ResizeToFit(width, height, pixelFormat, m_strideRequirement);
    RawBitmap srcBitmap(pContent, width, height, pixelFormat, m_stride);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }


  Bitmap::Bitmap(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride)
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(pixelFormat)
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
    if (pContent == nullptr)
      throw std::invalid_argument("pContent can not be null");
    if (std::abs(stride) < PixelFormatUtil::CalcMinimumStride(pixelFormat, width))
      throw std::invalid_argument("stride is smaller than the width allows");

    ResizeToFit(width, height, pixelFormat, m_strideRequirement);
    RawBitmap srcBitmap(pContent, width, height, pixelFormat, stride);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }


  Bitmap::Bitmap(const RawBitmap& srcBitmap)
    : m_content()
    , m_width(0)
    , m_height(0)
    , m_stride(0)
    , m_pixelFormat(srcBitmap.GetPixelFormat())
    , m_strideRequirement(StrideRequirement::Any)
    , m_bIsLocked(false)
  {
    if (srcBitmap.Content() == nullptr)
      throw std::invalid_argument("Content can not be null");

    ResizeToFit(srcBitmap.Width(), srcBitmap.Height(), srcBitmap.GetPixelFormat(), m_strideRequirement);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }


  int32_t Bitmap::Width() const
  {
    return m_width;
  }


  int32_t Bitmap::Height() const
  {
    return m_height;
  }


  int32_t Bitmap::Stride() const
  {
    return m_stride;
  }


  PixelFormat::Enum Bitmap::GetPixelFormat() const
  {
    return m_pixelFormat;
  }


  StrideRequirement::Enum Bitmap::GetStrideRequirement() const
  {
    return m_strideRequirement;
  }


  int32_t Bitmap::GetPreferredStride(const PixelFormat::Enum pixelFormat) const
  {
    return (pixelFormat == m_pixelFormat ? m_stride : PixelFormatUtil::CalcMinimumStride(m_width, pixelFormat, m_strideRequirement));
  }


  void Bitmap::Reset(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const BitmapClearMethod::Enum clearMethod)
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is locked");
    
    if (width != m_width || height != m_height || pixelFormat != m_pixelFormat)
    {
      ResizeToFit(width, height, pixelFormat, StrideRequirement::Any);
    }
    if (clearMethod != BitmapClearMethod::DontModify)
      Clear(clearMethod);
  }


  void Bitmap::Reset(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride, const BitmapClearMethod::Enum clearMethod)
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is locked");
    if (width != m_width || height != m_height || pixelFormat != m_pixelFormat || stride != m_stride)
    {
      if (stride < 0)
        throw std::invalid_argument("stride can not be < 0");

      ResizeToFit(width, height, pixelFormat, StrideRequirement::Any, stride);
    }
    if (clearMethod != BitmapClearMethod::DontModify)
      Clear(clearMethod);
  }


  void Bitmap::Reset(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat)
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is locked");
    if (pContent == nullptr)
      throw std::invalid_argument("pContent can not be null");

    ResizeToFit(width, height, pixelFormat, StrideRequirement::Any);
    RawBitmap srcBitmap(pContent, width, height, pixelFormat);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }


  void Bitmap::Reset(const void*const pContent, const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const int32_t stride)
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is locked");
    if (pContent == nullptr)
      throw std::invalid_argument("pContent can not be null");
    if (std::abs(stride) < PixelFormatUtil::CalcMinimumStride(pixelFormat, width))
      throw std::invalid_argument("stride is smaller than the width allows");

    ResizeToFit(width, height, pixelFormat, StrideRequirement::Any);
    RawBitmap srcBitmap(pContent, width, height, pixelFormat, stride);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }


  void Bitmap::Reset(const RawBitmap& srcBitmap)
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is locked");
    if (!srcBitmap.IsValid())
      throw std::invalid_argument("invalid srcBitmap");

    Reset(srcBitmap.Width(), srcBitmap.Height(), srcBitmap.GetPixelFormat(), BitmapClearMethod::Clear);
    RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
    Copy(dstBitmap, srcBitmap);
  }

  
  void Bitmap::SetNativePixel(const int x, const int y, const uint32_t color)
  {
    if (x < 0 || y < 0 || x >= m_width || y >= m_height)
      return;

    uint8_t* pDst = reinterpret_cast<uint8_t*>(m_content.data());
    switch (m_pixelFormat)
    {
    case PixelFormat::ALPHA8:
    case PixelFormat::LUMINANCE8:
      pDst[x + (y*m_stride)] = static_cast<uint8_t>(color & 0xFF);
      break;
    case PixelFormat::RGB888:
    case PixelFormat::BGR888:
      pDst += (x * 3) + (y*m_stride);
      pDst[0] = static_cast<uint8_t>(color & 0xFF);
      pDst[1] = static_cast<uint8_t>((color >> 8) & 0xFF);
      pDst[2] = static_cast<uint8_t>((color >> 16) & 0xFF);
      break;
    case PixelFormat::RGBA8888:
    case PixelFormat::BGRA8888:
      pDst += (x * 4) + (y*m_stride);
      pDst[0] = static_cast<uint8_t>(color & 0xFF);
      pDst[1] = static_cast<uint8_t>((color >> 8) & 0xFF);
      pDst[2] = static_cast<uint8_t>((color >> 16) & 0xFF);
      pDst[3] = static_cast<uint8_t>((color >> 24) & 0xFF);
      break;
    default:
      throw UnsupportedPixelFormatException(m_pixelFormat);
    }
  }


  uint32_t Bitmap::GetNativePixel(const int x, const int y) const
  {
    if (x < 0 || y < 0 || x >= m_width || y >= m_height)
      return 0;

    const uint8_t*pSrc = reinterpret_cast<const uint8_t*>(m_content.data());
    switch (m_pixelFormat)
    {
    case PixelFormat::ALPHA8:
    case PixelFormat::LUMINANCE8:
      return pSrc[x + (y*m_stride)];
    case PixelFormat::RGB888:
    case PixelFormat::BGR888:
      pSrc += (x * 3) + (y*m_stride);
      return static_cast<uint32_t>(pSrc[0]) | (static_cast<uint32_t>(pSrc[1]) << 8) | (static_cast<uint32_t>(pSrc[2]) << 16);
    case PixelFormat::RGBA8888:
    case PixelFormat::BGRA8888:
      pSrc += (x * 4) + (y*m_stride);
      return static_cast<uint32_t>(pSrc[0]) | (static_cast<uint32_t>(pSrc[1]) << 8) | (static_cast<uint32_t>(pSrc[2]) << 16) | (static_cast<uint32_t>(pSrc[3]) << 24);
    default:
      throw UnsupportedPixelFormatException(m_pixelFormat);
    }
  }


  RawBitmap Bitmap::Lock() const
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is already locked");
    m_bIsLocked = true;
    return RawBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
  }


  RawBitmapEx Bitmap::LockEx()
  {
    if (m_bIsLocked)
      throw UsageErrorException("The bitmap is already locked");
    m_bIsLocked = true;
    return RawBitmapEx(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
  }


  void Bitmap::Unlock()
  {
    if (!m_bIsLocked)
      throw UsageErrorException("The bitmap is not locked");
    m_bIsLocked = false;
  }


  void Bitmap::Unlock() const
  {
    if (!m_bIsLocked)
      throw UsageErrorException("The bitmap is not locked");
    m_bIsLocked = false;
  }


  void Bitmap::ResizeToFit(const int32_t width, const int32_t height, const PixelFormat::Enum pixelFormat, const StrideRequirement::Enum strideRequirement, const int32_t stride)
  {
    if (width < 0)
      throw std::invalid_argument("width can not be negative");
    if (height < 0)
      throw std::invalid_argument("height can not be negative");

    const int32_t bytesPerPixel = PixelFormatUtil::GetBytesPerPixel(pixelFormat);
    const int32_t minStride = PixelFormatUtil::CalcMinimumStride(width, bytesPerPixel, strideRequirement);
    int32_t chosenStride;
    
    if (stride == 0)
    {
      chosenStride = minStride;
    }
    else
    {
      if( ! PixelFormatUtil::IsValidStride(width, bytesPerPixel, strideRequirement, stride) )
        throw std::invalid_argument("Supplied stride does not obey the StrideRequirement");

      chosenStride = stride;
    }

    const int32_t totalByteSize = (height * chosenStride);
    const int32_t uint32Size = (totalByteSize / sizeof(uint32_t)) + ((totalByteSize % sizeof(uint32_t)) != 0 ? 1 : 0);
    if (m_content.size() != size_t(uint32Size))
      m_content.resize(uint32Size);

    // Update the members
    m_width = width;
    m_height = height;
    m_stride = chosenStride;
    m_pixelFormat = pixelFormat;
    m_strideRequirement = strideRequirement;
  }


  void Bitmap::Clear(const BitmapClearMethod::Enum clearMethod)
  {
    switch (clearMethod)
    {
    case BitmapClearMethod::Clear:
      std::fill(m_content.begin(), m_content.end(), 0);
      break;
    case BitmapClearMethod::DontClear:
      if (PixelFormatUtil::CalcMinimumStride(m_width, m_pixelFormat) != m_stride)
      {
        RawBitmapEx dstBitmap(m_content.data(), m_width, m_height, m_pixelFormat, m_stride);
        RawBitmapUtil::ClearPadding(dstBitmap);
      }
      break;
    case BitmapClearMethod::DontModify:
      break;
    default:
      throw NotSupportedException("Unknown BitmapClearMethod");
    }

  }

}
