#ifdef __ANDROID__
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <iostream>
#include <android/sensor.h>
#include <android/log.h>
#include <android_native_app_glue.h>
#include <android/native_window_jni.h>
#include <FslBase/Log/Log.hpp>
#include <FslGraphics/Window/INativeWindowEventQueue.hpp>
#include <FslGraphics/Window/NativeWindowEventHelper.hpp>
#include <FslGraphicsEGL/Exceptions.hpp>
#include <FslGraphicsEGL/Window/Android/EGLNativeWindowSystemAndroidTag.hpp>
#include "EGLNativeWindowAndroid.hpp"

#if 0
#define LOCAL_LOG(X) FSLLOG("EGLNativeWindowAndroid: " << X)
#else
#define LOCAL_LOG(X) {}
#endif

namespace Fsl
{

  namespace
  {
    std::weak_ptr<INativeWindowEventQueue> g_eventQueue;

    EGLNativeWindowType g_hWindow = nullptr;
    bool g_isSuspended = false;
    bool g_isActivated = false;


    void ShowUI(android_app* pAppState)
    {
      JNIEnv *jni;
      pAppState->activity->vm->AttachCurrentThread( &jni, nullptr );

      //Default class retrieval
      jclass clazz = jni->GetObjectClass( pAppState->activity->clazz );
      jmethodID methodID = jni->GetMethodID( clazz, "showUI", "()V" );
      jni->CallVoidMethod( pAppState->activity->clazz, methodID );

      pAppState->activity->vm->DetachCurrentThread();
    }


    void PostActivated(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, bool activate)
    {
      if( activate != g_isActivated )
      {
        g_isActivated = activate;
        eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowActivationEvent(activate));
      }
    }


    void PostSuspend(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, bool suspend)
    {
      if( suspend != g_isSuspended )
      {
        g_isSuspended = suspend;
        if( suspend && g_isActivated)
          PostActivated(eventQueue, false);

        eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowSuspendEvent(suspend));
      }
    }


    void CmdInitWindow(android_app* pAppState, const std::shared_ptr<INativeWindowEventQueue>& eventQueue)
    {
      // The window is being shown, get it ready.
      if (pAppState->window != nullptr)
      {
        LOCAL_LOG("NativeWindow init");
        PostSuspend(eventQueue, false);
        g_hWindow = pAppState->window;
      }
      else
      {
        g_hWindow = nullptr;
        LOCAL_LOG("No window pointer");
      }
    }


    void CmdShutdownWindow(const std::shared_ptr<INativeWindowEventQueue>& eventQueue)
    {
      // The window is being hidden or closed, clean it up.
      PostSuspend(eventQueue, true);
      g_hWindow = nullptr;
    }


    void CmdSetFocus(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, const bool bEnabled)
    {
      PostActivated(eventQueue, bEnabled);
    }


    void CmdLowMemory(const std::shared_ptr<INativeWindowEventQueue>& eventQueue)
    {
      eventQueue->PostEvent(NativeWindowEventHelper::EncodeLowMemoryEvent());
    }


    void CmdCatcher(android_app* pAppState, int32_t cmd )
    {
      LOCAL_LOG("HandleCmd");

      std::shared_ptr<INativeWindowEventQueue> eventQueue = g_eventQueue.lock();
      if( ! eventQueue )
        return;

      switch (cmd)
      {
      case APP_CMD_INPUT_CHANGED:
        LOCAL_LOG("APP_CMD_INPUT_CHANGED");
        break;
      case APP_CMD_INIT_WINDOW:
        LOCAL_LOG("APP_CMD_INIT_WINDOW");
        CmdInitWindow(pAppState, eventQueue);
        break;
      case APP_CMD_TERM_WINDOW:
        LOCAL_LOG("APP_CMD_TERM_WINDOW");
        CmdShutdownWindow(eventQueue);
        break;
      case APP_CMD_WINDOW_RESIZED:
        LOCAL_LOG("APP_CMD_WINDOW_RESIZED");
        break;
      case APP_CMD_WINDOW_REDRAW_NEEDED:
        LOCAL_LOG("APP_CMD_WINDOW_RESIZED");
        break;
      case APP_CMD_CONTENT_RECT_CHANGED:
        LOCAL_LOG("APP_CMD_CONTENT_RECT_CHANGED");
        break;

      case APP_CMD_GAINED_FOCUS:
        LOCAL_LOG("APP_CMD_GAINED_FOCUS");
        CmdSetFocus(eventQueue, true);
        break;
      case APP_CMD_LOST_FOCUS:
        LOCAL_LOG("APP_CMD_LOST_FOCUS");
        CmdSetFocus(eventQueue, false);
        break;

      case APP_CMD_CONFIG_CHANGED:
        LOCAL_LOG("APP_CMD_CONFIG_CHANGED");
        break;
      case APP_CMD_LOW_MEMORY:
        LOCAL_LOG("APP_CMD_LOW_MEMORY");
        CmdLowMemory(eventQueue);
        break;

      case APP_CMD_START:
        LOCAL_LOG("APP_CMD_START");
        break;
      case APP_CMD_RESUME:
        LOCAL_LOG("APP_CMD_RESUME");
        break;
      case APP_CMD_SAVE_STATE:
        LOCAL_LOG("APP_CMD_SAVE_STATE");
        break;
      case APP_CMD_PAUSE:
        LOCAL_LOG("APP_CMD_PAUSE");
        break;
      case APP_CMD_STOP:
        LOCAL_LOG("APP_CMD_STOP");
        //PostSuspend(eventQueue, true);
        break;
      case APP_CMD_DESTROY:
        LOCAL_LOG("APP_CMD_DESTROY");
        break;
      default:
        LOCAL_LOG("Unknown cmd: " << cmd);
        break;
      }
    }


    int32_t InputCatcher(android_app* pAppState, struct AInputEvent* event )
    {
      //EGLNativeWindowAndroid* host = static_cast<EGLNativeWindowAndroid*>(pAppState->userData);
      //if( host == nullptr )
      //{
      //  LOCAL_LOG("ERROR: No host configured (cmd ignored)");
      //  return 0;
      //}
      return 0;
    }



    bool LocalProcessMessages(android_app* pAppState)
    {
      // Read all pending events.
      int id;
      int events;
      android_poll_source* source;

      // We loop until all events are read, then continue to draw the next frame of animation.
      while ((id = ALooper_pollAll(0, nullptr, &events, (void**)&source)) >= 0)
      {
        // Process this event.
        if (source != nullptr)
        {
          source->process(pAppState, source);
        }

        //          g_engine.ProcessSensors( id );

        // Check if we are exiting.
        if (pAppState->destroyRequested != 0)
        {
          LOCAL_LOG("Destroy requested");
          return false;
        }
      }
      return true;
    }
  }


  EGLNativeWindowSystemAndroid::EGLNativeWindowSystemAndroid(const NativeWindowSystemSetup& setup)
    : m_pAppState()
  {
    g_eventQueue = setup.GetEventQueue();

    const std::shared_ptr<EGLNativeWindowSystemAndroidTag> tag = setup.GetTag<EGLNativeWindowSystemAndroidTag>();
    m_pAppState = tag->State;

    // Configure the callback handlers
    m_pAppState->onAppCmd = CmdCatcher;
    m_pAppState->onInputEvent = InputCatcher;
  }


  EGLNativeWindowSystemAndroid::~EGLNativeWindowSystemAndroid()
  {
    g_eventQueue.reset();
    m_pAppState->onAppCmd = nullptr;
    m_pAppState->onInputEvent = nullptr;
  }


  EGLNativeDisplayType EGLNativeWindowSystemAndroid::GetDisplayType() const
  {
    return EGL_DEFAULT_DISPLAY;
  }


  std::shared_ptr<IEGLNativeWindow> EGLNativeWindowSystemAndroid::CreateNativeWindow(const NativeWindowSetup& nativeWindowSetup, const NativeEGLSetup& nativeEglSetup)
  {
    return std::shared_ptr<IEGLNativeWindow>(new EGLNativeWindowAndroid(nativeWindowSetup, nativeEglSetup, m_pAppState));
  }


  bool EGLNativeWindowSystemAndroid::ProcessMessages(const NativeWindowProcessMessagesArgs& args)
  {
    return LocalProcessMessages(m_pAppState);
  }


  EGLNativeWindowAndroid::EGLNativeWindowAndroid(const NativeWindowSetup& nativeWindowSetup, const NativeEGLSetup& nativeEglSetup, android_app* pAppState)
    : m_eventQueue(nativeWindowSetup.GetEventQueue())
    , m_hDisplay(nativeEglSetup.Display)
    , m_eglConfig(nativeEglSetup.Config)
    , m_pAppState(pAppState)
    , m_hWindow(nullptr)
  {
    // Configure the callback handlers
    m_pAppState->userData = this;

    // Wait for a window to be ready
    if( g_hWindow == nullptr )
    {
      LOCAL_LOG("Waiting for window to be ready");
      WaitForWindowReady();
    }
    LOCAL_LOG("Window created");
    m_hWindow = g_hWindow;


    LOCAL_LOG("NativeWindow ANativeWindow_setBuffersGeometry");

    // EGL_NATIVE_VISUAL_ID is an attribute of the EGLConfig that is guaranteed to be accepted by ANativeWindow_setBuffersGeometry().
    // As soon as we picked a EGLConfig, we can safely reconfigure the ANativeWindow buffers to match, using EGL_NATIVE_VISUAL_ID.
    EGLint format;
    eglGetConfigAttrib( m_hDisplay, m_eglConfig, EGL_NATIVE_VISUAL_ID, &format );
    ANativeWindow_setBuffersGeometry( pAppState->window, 0, 0, format );

    LOCAL_LOG("NativeWindow show UI");
    ShowUI(pAppState);
    LOCAL_LOG("NativeWindow ready");
  }


  EGLNativeWindowAndroid::~EGLNativeWindowAndroid()
  {
    // Clear the pointer
    m_pAppState->userData = nullptr;
  }


  EGLNativeWindowType EGLNativeWindowAndroid::GetWindowType() const
  {
    return m_hWindow;
  }


  void EGLNativeWindowAndroid::WaitForWindowReady()
  {
    bool bContinue;
    // Keep processing messages until we get a active window or a termination request
    while( (bContinue = LocalProcessMessages(m_pAppState)) == true && g_hWindow == nullptr )
    {
    }

    if( ! bContinue )
      throw EGLGraphicsException("Failed to create window");
  }
}
#endif
