#ifndef FSLGRAPHICS_DRAFT_PROCEDURAL_GENERICMESH_HPP
#define FSLGRAPHICS_DRAFT_PROCEDURAL_GENERICMESH_HPP
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics_Draft/Procedural/GenericMesh_fwd.hpp>
#include <FslGraphics_Draft/Procedural/IndexUtil.hpp>
#include <FslGraphics/PrimitiveTypeUtil.hpp>
#include <FslGraphics/Exceptions.hpp>
#include <algorithm>

namespace Fsl {
namespace Procedural
{
  // The class is defined in GenericMesh_fwd.hpp and this file contains the implementation of the 'templated' functions of GenericMesh.
  // This allows us to use templates without 'spreading' dependencies to files that don't need it by allowing them to only include the _fwd file

  template<typename TVertex, typename TIndex>
  GenericMesh<TVertex, TIndex>::GenericMesh()
    : m_vertices()
    , m_indices()
    , m_primitiveType(PrimitiveType::LineList)
    , m_primitiveCount(0)
  {
  }


  template<typename TVertex, typename TIndex>
  GenericMesh<TVertex, TIndex>::GenericMesh(const TVertex*const pVertices, const std::size_t vertexCount, const TIndex*const pIndices, const std::size_t indexCount, const PrimitiveType::Enum primitiveType)
    : m_vertices(pVertices, pVertices + vertexCount)
    , m_indices(pIndices, pIndices + indexCount)
    , m_primitiveType(primitiveType)
    , m_primitiveCount(PrimitiveTypeUtil::CalcPrimitiveCount(indexCount, primitiveType))
  {

  }


  template<typename TVertex, typename TIndex>
  GenericMesh<TVertex, TIndex>::GenericMesh(const std::vector<TVertex>& vertices, const std::vector<TIndex>& indices, const PrimitiveType::Enum primitiveType)
    : m_vertices(vertices)
    , m_indices(indices)
    , m_primitiveType(primitiveType)
    , m_primitiveCount(PrimitiveTypeUtil::CalcPrimitiveCount(indices.size(), primitiveType))
  {
  }


  template<typename TVertex, typename TIndex>
  GenericMesh<TVertex, TIndex>::GenericMesh(const std::vector<TVertex>& vertices, const int vertexStartOffset, const int vertexCount,
                                            const std::vector<TIndex>& indices, const int indexStartOffset, const int indexCount,
                                            const PrimitiveType::Enum primitiveType)
    : m_vertices(std::max(vertexCount, 0))
    , m_indices(std::max(indexCount, 0))
    , m_primitiveType(primitiveType)
    , m_primitiveCount(0)
  {
    if (vertexStartOffset < 0 || vertexCount < 0 || std::size_t(vertexStartOffset + vertexCount) > vertices.size() ||
        indexStartOffset < 0 || indexCount < 0 || std::size_t(indexStartOffset + indexCount) > indices.size())
    {
      throw std::invalid_argument("One of the arguments is invalid");
    }

    std::copy(vertices.begin() + vertexStartOffset, vertices.begin() + (vertexStartOffset + vertexCount), m_vertices.begin());
    std::copy(indices.begin() + indexStartOffset, indices.begin() + (indexStartOffset + indexCount), m_indices.begin());

    m_primitiveCount = PrimitiveTypeUtil::CalcPrimitiveCount(indexCount, primitiveType);
  }


  template<typename TVertex, typename TIndex>
  void GenericMesh<TVertex, TIndex>::Reset()
  {
    m_vertices.clear();
    m_indices.clear();
    m_primitiveType = PrimitiveType::LineList;
    m_primitiveCount = 0;
  }


  template<typename TVertex, typename TIndex>
  void GenericMesh<TVertex, TIndex>::Reset(const std::vector<TVertex>& vertices, const std::vector<TIndex>& indices, const PrimitiveType::Enum primitiveType)
  {
    m_vertices.resize(vertices.size());
    m_indices.resize(indices.size());

    // Copy the input
    std::copy(vertices.begin(), vertices.end(), m_vertices.begin());
    std::copy(indices.begin(), indices.end(), indices.begin());

    m_primitiveType = primitiveType;
    m_primitiveCount = PrimitiveTypeUtil::CalcPrimitiveCount(indices.size(), primitiveType);
  }


  template<typename TVertex, typename TIndex>
  void GenericMesh<TVertex, TIndex>::Reset(const std::vector<TVertex>& vertices, const int vertexStartOffset, const int vertexCount,
                                           const std::vector<TIndex>& indices, const int indexStartOffset, const int indexCount,
                                           const PrimitiveType::Enum primitiveType)
  {
    if (vertexStartOffset < 0 || vertexCount < 0 || std::size_t(vertexStartOffset + vertexCount) > vertices.size() ||
        indexStartOffset < 0 || indexCount < 0 || std::size_t(indexStartOffset + indexCount) > indices.size())
    {
      throw std::invalid_argument("One of the arguments is invalid");
    }

    m_vertices.resize(vertexCount);
    m_indices.resize(indexCount);

    std::copy(vertices.begin() + vertexStartOffset, vertices.begin() + (vertexStartOffset + vertexCount), m_vertices.begin());
    std::copy(indices.begin() + indexStartOffset, indices.begin() + (indexStartOffset + indexCount), m_indices.begin());

    m_primitiveCount = PrimitiveTypeUtil::CalcPrimitiveCount(indexCount, primitiveType);
  }


  template<typename TVertex, typename TIndex>
  PrimitiveType::Enum GenericMesh<TVertex, TIndex>::GetPrimitiveType() const
  {
    return m_primitiveType;
  }


  template<typename TVertex, typename TIndex>
  int GenericMesh<TVertex, TIndex>::GetVertexCount() const
  {
    return m_vertices.size();
  }


  template<typename TVertex, typename TIndex>
  int GenericMesh<TVertex, TIndex>::GetIndexCount() const
  {
    return m_indices.size();
  }


  template<typename TVertex, typename TIndex>
  int GenericMesh<TVertex, TIndex>::GetPrimitiveCount() const
  {
    return m_primitiveCount;
  }


  template<typename TVertex, typename TIndex>
  const TVertex* GenericMesh<TVertex, TIndex>::GetVertices() const
  {
    return m_vertices.data();
  }


  template<typename TVertex, typename TIndex>
  const TIndex* GenericMesh<TVertex, TIndex>::GetIndices() const
  {
    return m_indices.data();
  }


  template<typename TVertex, typename TIndex>
  const std::vector<TVertex>& GenericMesh<TVertex, TIndex>::GetVertexArray() const
  {
    return m_vertices;
  }


  template<typename TVertex, typename TIndex>
  const std::vector<TIndex>& GenericMesh<TVertex, TIndex>::GetIndexArray() const
  {
    return m_indices;
  }

}}

#endif


